""" 
Essentially data scripts for handling periodic table properties.
Needed for handling few cases not readily handled elswhere and for architector-specific tunings.

Developed by Michael Taylor and Dan Burril.
"""

import numpy as np
import copy 

elements = ('X',
        "H",                                                                                                 "He",
        "Li", "Be",                                                              "B",  "C",  "N",  "O",  "F", "Ne",
        "Na", "Mg",                                                             "Al", "Si",  "P",  "S", "Cl", "Ar",
        "K", "Ca", "Sc", "Ti",  "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr",
        "Rb", "Sr",  "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te",  "I", "Xe",
        "Cs", "Ba", 
                    "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu",
                        "Hf", "Ta",  "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn",
        "Fr", "Ra",
                    "Ac", "Th", "Pa",  "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr",
                        "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"
        )  

# Data from Dalton Trans., 2013, 42, 8617-8636 -> Assigning Pm to average of neighbors = 2.92, Fr to 3.28, Ra to 3.10
rvdw = (0.00,
        1.20,                                                                                                 1.43,
        2.12, 1.98,                                                             1.91, 1.77, 1.66, 1.50, 1.46, 1.58,
        2.50, 2.51,                                                             2.25, 2.19, 1.90, 1.89, 1.82, 1.83,
        2.73, 2.62, 2.58, 2.46, 2.42, 2.45, 2.45, 2.44, 2.40, 2.40, 2.38, 2.39, 2.32, 2.29, 1.88, 1.82, 1.86, 2.25,
        3.21, 2.84, 2.75, 2.52, 2.56, 2.45, 2.44, 2.46, 2.44, 2.15, 2.53, 2.49, 2.43, 2.42, 2.47, 1.99, 2.04, 2.06,
        3.48, 3.03, 
                    2.98, 2.88, 2.92, 2.95, 2.92 , 2.90, 2.87, 2.83, 2.79, 2.87, 2.81, 2.83, 2.79, 2.80, 2.74,
                        2.63, 2.53, 2.57, 2.49, 2.48, 2.41, 2.29, 2.32, 2.45, 2.47, 2.60, 2.54, "Po", "At", "Rn",
        3.28, 3.10,
                    2.80, 2.93, 2.88, 2.71, 2.82, 2.81, 2.83, 3.05, 3.40, 3.05, 2.70, "Fm", "Md", "No", "Lr",
                        "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"
        )


# Data from J. Phys. Chem. A 2015, 119, 2326-2337
# single bond only
rcov1 = (0.00,
        0.32,                                                                                                 0.46,
        1.33, 1.02,                                                             0.85, 0.75, 0.71, 0.63, 0.64, 0.67,
        1.55, 1.39,                                                             1.26, 1.16, 1.11, 1.03, 0.99, 0.96,
        1.96, 1.71, 1.48, 1.36, 1.34, 1.22, 1.19, 1.16, 1.11, 1.10, 1.12, 1.18, 1.24, 1.21, 1.21, 1.16, 1.14, 1.17,
        2.10, 1.85, 1.63, 1.54, 1.47, 1.38, 1.28, 1.25, 1.25, 1.20, 1.28, 1.36, 1.42, 1.40, 1.40, 1.36, 1.33, 1.31,
        2.32, 1.96, 
                    1.80, 1.63, 1.76, 1.74, 1.73, 1.72, 1.68, 1.69, 1.68, 1.67, 1.66, 1.65, 1.64, 1.70, 1.62,
                        1.52, 1.46, 1.37, 1.31, 1.29, 1.22, 1.23, 1.24, 1.33, 1.44, 1.44, 1.51, 1.45, 1.47, 1.42,
        2.23, 2.01,
                    1.86, 1.75, 1.69, 1.70, 1.71, 1.72, 1.66, 1.66, 1.68, 1.68, 1.65, 1.67, 1.73, 1.76, 1.61,
                        1.57, 1.49, 1.43, 1.41, 1.34, 1.29, 1.28, 1.21, 1.22, 1.36, 1.43, 1.62, 1.75, 1.65, 1.57
)

# Used mostly for charge on ligand detection. Metal valence electron counts shouldn't be referenced anywhere (untrustworthy)
valence_electrons = (0.00,
    1,                                                    2,
    1, 2,                                  3, 4, 5, 6, 7, 8,
    1, 2,                                  3, 4, 5, 6, 7, 8,
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 3, 4, 5, 6, 7, 8,
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 3, 4, 5, 6, 7, 8,
    1, 2, 
                3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 3, # Lu - 3 valence elctrons
                    4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, # Hg - 2 valence electrons
    1, 2,
                3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 3, # Lr - 3 valence elctrons
                    4, 5, 6, 7, 8, 9, 10, 11, 2, 3, 4, 5, 6, 7, 8, # Cn - 2 valence electrons
)

masses = (0.00, # Masses
    1.0080,                                                                                                                                                                   4.0026,
    6.9400,   9.0122,                                                                                                      10.8100,  12.0110,  14.0070,  15.9990,  18.9980,  20.1800,
    22.9900,  24.3050,                                                                                                      26.9820,  28.0850,  30.9740,  32.0600,  35.4500,  39.9480,
    39.0980,  40.0780,  44.9560,  47.8670,  50.9420,  51.9960,  54.9380,  55.8450,  58.9330,  58.6930,  63.5460,  65.3800,  69.7230,  72.6300,  74.9220,  78.9710,  79.9040,  83.7980,
    85.4680,  87.6200,  88.9060,  91.2240,  92.9060,  95.9500,  97.0000, 101.0700, 102.9100, 106.4200, 107.8700, 112.4100, 114.8200, 118.7100, 121.7600, 127.6000, 126.9000, 131.2900,
    132.9100, 137.3300, 
                        138.9100, 140.1200, 140.9100, 144.2400, 145.0000, 150.3600, 151.9600, 157.2500, 158.9300, 162.5000, 164.9300, 167.2600, 168.9300, 173.0500, 174.9700,
                                178.4900, 180.9500, 183.8400, 186.2100, 190.2300, 192.2200, 195.0800, 196.9700, 200.5900, 204.3800, 207.2000, 208.9800, 209.0000, 210.0000, 222.0000,
    223.0000, 226.0000,
                        227.0000, 232.0400, 231.0400, 238.0300, 237.0000, 244.0000, 243.0000, 247.0000, 247.0000, 251.0000, 252.0000, 257.0000, 258.0000, 259.0000, 266.0000,
                                267.0000, 268.0000, 269.0000, 270.0000, 269.0000, 278.0000, 281.0000, 282.0000, 285.0000, 286.0000, 289.0000, 289.0000, 293.0000, 294.0000, 294.0000
)

# Filled valence numbers
filled_valence_electrons=(2,8,18,32)

# XTB/GFN-2 limitation - Z <= 86
limited_rcov1 = rcov1[0:87]
limited_elements = elements[0:87]
limited_rvdw = rvdw[0:87]
       
lanthanides = ['La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy',
                    'Ho', 'Er', 'Tm', 'Yb', 'Lu']

limited_lanthanides = [x for x in lanthanides if x in limited_elements]

actinides = ['Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf',
                    'Es', 'Fm', 'Md', 'No', 'Lr']

limited_actinides = [x for x in actinides if x in limited_elements]

transition_metals = [
            'Sc', 'Ti', 'V',  'Cr', 'Mn', 'Fe', 'Co',  'Ni', 'Cu',  'Zn', 
            'Y',  'Zr', 'Nb',  'Mo', 'Tc',  'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 
            'Hf',  'Ta', 'W',  'Re',  'Os',  'Ir', 'Pt',  'Au', 'Hg',
            'Rf', 'Db', 'Sg', 'Bh', 'Hs'
            ]

limited_transition_metals = [x for x in transition_metals if x in limited_elements]

post_transition_metals = [ 
            'Al',
            'Ga',
            'In', 'Sn',
            'Tl', 'Pb', 'Bi', 
            'Nh', 'Fl', 'Mc', 'Lv'
        ]

limited_post_transition_metals = [x for x in post_transition_metals if x in limited_elements]

alkali_and_alkaline_earth = [
            'Li', 'Be', 
            'Na', 'Mg', 
            'K', 'Ca', 
            'Rb', 'Sr', 
            'Cs', 'Ba', 
            'Fr', 'Ra'
            ]

alkali_metals = ['Li','Na','K','Rb','Cs','Fr']

alkaline_earth_metals = ['Be','Mg','Ca','Sr','Ba','Ra']

metalloids = ['B','Si','Ge','As','Sb','Te']

limited_alkali_and_alkaline_earth = [x for x in alkali_and_alkaline_earth if x in limited_elements]

heavy_metals = lanthanides + actinides

limited_heavy_metals = limited_lanthanides + limited_actinides

all_metals = lanthanides + actinides + alkali_and_alkaline_earth + post_transition_metals + transition_metals

limited_all_metals = limited_lanthanides + limited_actinides + limited_alkali_and_alkaline_earth \
    + limited_post_transition_metals + limited_transition_metals

# Dictionary for common charges of metals
metal_charge_dict = {
    # Lanthanides
    'La': 3, 'Ce': 3, 'Pr': 3, 'Nd': 3, 'Pm': 3, 'Sm': 3, 'Eu': 3, 'Gd': 3, 'Tb': 3, 'Dy': 3,
    'Ho': 3, 'Er': 3, 'Tm': 3, 'Yb': 3, 'Lu': 3,
    # Actinides
    'Ac': 3, 'Th': 4, 'Pa': 5, 'U': 4, 'Np': 4, 'Pu': 4, 'Am': 3, 'Cm': 3, 'Bk': 3, 'Cf': 3,
    'Es': 3, 'Fm': 3, 'Md': 3, 'No': 2, 'Lr': 3,
    # First row transition metals 
    'Sc': 3, 'Ti': 4, 'V': 5,  'Cr': 3, 'Mn': 2, 'Fe': 2, 'Co': 2, 'Ni': 2, 'Cu': 2, 'Zn': 2, 
    # Second row transition metals 
    'Y': 3, 'Zr': 4, 'Nb': 5,  'Mo': 6, 'Tc': 5, 'Ru': 2, 'Rh': 1, 'Pd': 2, 'Ag': 1, 'Cd': 2, 
    # Third row transition metals
    'Hf': 4, 'Ta': 5, 'W': 6,  'Re': 4, 'Os': 2, 'Ir': 3, 'Pt': 2, 'Au': 3, 'Hg': 2,
    # 4th row transition metals
    'Rf': 3, 'Db': 5, 'Sg': 6, 'Bh': 7, 'Hs': 8,
    # Post-transition metals
    'Al': 3,
    'Ga': 3,
    'In': 3, 'Sn': 4,
    'Tl': 3, 'Pb': 2, 'Bi': 3, 
    'Nh': 1, 'Fl': 2, 'Mc': 1, 'Lv': 2,
    # Alkali/earth metals
    'Li': 1, 'Be': 2 , 
    'Na': 1, 'Mg': 2, 
    'K': 1, 'Ca': 2, 
    'Rb': 1, 'Sr': 2, 
    'Cs': 1, 'Ba': 2, 
    'Fr': 1, 'Ra': 2
    }

# Dictionary for common spin (alpha-beta) of metals:
metal_spin_dict = {
    # Lanthanides
    'La': 0, 'Ce': 1, 'Pr': 2, 'Nd': 3, 'Pm': 4, 'Sm': 5, 'Eu': 6, 'Gd': 7, 'Tb': 6, 'Dy': 5,
    'Ho': 4, 'Er': 3, 'Tm': 2, 'Yb': 1, 'Lu': 0,
    # Actinides
    'Ac': 0, 'Th': 0, 'Pa': 0, 'U': 2, 'Np': 3, 'Pu': 4, 'Am': 6, 'Cm': 7, 'Bk': 6, 'Cf': 5,
    'Es': 4, 'Fm': 3, 'Md': 2, 'No': 0, 'Lr': 0,
    # First row transition metals
    'Sc': 0, 'Ti': 0, 'V': 0,  'Cr': 3, 'Mn': 5, 'Fe': 4, 'Co': 3, 'Ni': 2, 'Cu': 1, 'Zn': 0, 
    # Second row transition metals 
    'Y': 0, 'Zr': 0, 'Nb': 0,  'Mo': 0, 'Tc': 2, 'Ru': 4, 'Rh': 2, 'Pd': 0, 'Ag': 0, 'Cd': 0, 
    # Third row transition metals
    'Hf': 0, 'Ta': 0, 'W': 0,  'Re': 1, 'Os': 0, 'Ir': 0, 'Pt': 0, 'Au': 2, 'Hg': 0,
    # 4th row transition metals
    'Rf': 1, 'Db': 0, 'Sg': 0, 'Bh': 0, 'Hs': 0,
    # Post-transition metals
    'Al': 0,
    'Ga': 0,
    'In': 0, 'Sn': 0,
    'Tl': 0, 'Pb': 0, 'Bi': 0, 
    'Nh': 0, 'Fl': 0, 'Mc': 2, 'Lv': 2,
    # Alkali/earth metals
    'Li': 0, 'Be': 0 , 
    'Na': 0, 'Mg': 0, 
    'K': 0, 'Ca': 0, 
    'Rb': 0, 'Sr': 0, 
    'Cs': 0, 'Ba': 0, 
    'Fr': 0, 'Ra': 0
    }

# Dictionary for common spin (alpha-beta) of metals:
second_choice_metal_spin_dict = {
    # Lanthanides
    #'La': 0, 'Ce': 1, -> No real alternative spins
    # Other lanthanides set to 1 lower spin state (nspin-=2)
    'Pr': 0, 'Nd': 1,'Pm': 2,'Sm': 3,'Eu': 4, 'Gd': 5,'Tb': 4, 'Dy': 3, 'Ho': 2, 'Er': 1, 'Tm': 0, 
    #'Yb': 1, 'Lu': 0, -> No real alternative spins
    # Actinides -> Right now actinides are being treated by the lanthanides anyways.
    # 'Ac': 0, 'Th': 0, 'Pa': 0, 'U': 2, 'Np': 2, 'Pu': 4, 'Am': 6, 'Cm': 7, 'Bk': 6, 'Cf': 5,
    # 'Es': 4, 'Fm': 3, 'Md': 2, 'No': 0, 'Lr': 0,
    # First row transition metals
    #'Sc': 0, -> No real alternative spins 
    # Other transition metals set to Low-spin/IS spin state (nspin-=2)
    'Ti': 0, 'V': 1, 'Cr': 0, 'Mn': 1, 'Fe': 0, 'Co': 1, 'Ni': 0, 'Cu': 2, 
    # 'Zn': 0, -> No real alternative spins
    # Second row transition metals 
    # 'Y': 0, -> No real alternative spins
    'Zr': 2, 'Nb': 2,  'Mo': 0, 'Tc': 1, 'Ru': 1, 'Rh': 0, 'Pd': 2, 'Ag': 2, 
    # 'Cd': 0, -> No real alternative spins
    # Third row transition metals
    #'Hf': 0, 'Ta': 0, -> No real alternative spins
    'W': 0,  'Re': 1, 'Os': 2, 'Ir': 2, 'Pt': 0, 'Au': 2, 'Hg': 3,
    # 4th row transition metals
    #'Rf': 1, 'Db': 0, -> No real alternative spins
    'Sg': 2, 'Bh': 2, 'Hs': 2,
    # Post-transition metals
    #'Al': 0,'Ga': 0, 'In': 0, -> No real alternative spins
    'Sn': 2,
    'Tl': 2, 'Pb': 2, 'Bi': 2, 
    'Nh': 2, 'Fl': 2, 'Mc': 0, 'Lv': 0,
    # Alkali/earth metals -> No real alternative spins
    # 'Li': 0, 'Be': 0 , 
    # 'Na': 0, 'Mg': 0, 
    # 'K': 0, 'Ca': 0, 
    # 'Rb': 0, 'Sr': 0, 
    # 'Cs': 0, 'Ba': 0, 
    # 'Fr': 0, 'Ra': 0
    }

metal_CN_dict = {
    # Lanthanides
    'La': [7,8,9], 'Ce': [7,8,9], 'Pr': [7,8,9], 'Nd': [7,8,9], 'Pm': [7,8,9], 'Sm': [7,8,9], 'Eu': [7,8,9], 'Gd': [7,8,9], 'Tb': [7,8,9], 'Dy': [7,8,9],
    'Ho': [7,8,9], 'Er': [7,8,9], 'Tm': [7,8,9], 'Yb': [7,8,9], 'Lu': [7,8,9],
    # Actinides
    'Ac': [6,7,8,9], 'Th': [6,7,8,9], 'Pa': [6,7,8,9], 'U': [6,7,8,9], 'Np': [6,7,8,9], 'Pu': [6,7,8,9], 'Am': [6,7,8,9], 'Cm': [6,7,8,9], 'Bk': [6,7,8,9], 'Cf': [6,7,8,9],
    'Es': [6,7,8,9], 'Fm': [6,7,8,9], 'Md': [6,7,8,9], 'No': [6,7,8,9], 'Lr': [6,7,8,9],
    # First row transition metals 
    'Sc': [4,6], 'Ti': [4,6], 'V': [4,6],  'Cr': [4,6], 'Mn': [4,6], 'Fe': [4,6], 'Co': [4,6], 'Ni': [4,6], 'Cu': [4,6], 'Zn': [4,6], 
    # Second row transition metals 
    'Y': [4,6], 'Zr': [4,6], 'Nb': [4,6],  'Mo': [4,6], 'Tc': [4,6], 'Ru': [4,6], 'Rh': [4,6], 'Pd': [4,6], 'Ag': [4,6], 'Cd': [4,6], 
    # Third row transition metals
    'Hf': [4,6], 'Ta': [4,6], 'W': [4,6],  'Re': [4,6], 'Os': [4,6], 'Ir': [4,6], 'Pt': [4,6], 'Au': [4,6], 'Hg': [4,6],
    # 4th row transition metals
    'Rf': [4,6], 'Db': [4,6], 'Sg': [4,6], 'Bh': [4,6], 'Hs': [4,6],
    # Post-transition metals
    'Al': [4,6],
    'Ga': [4,6],
    'In': [4,6], 'Sn': [4,6],
    'Tl': [4,6], 'Pb': [4,6], 'Bi': [4,6], 
    'Nh': [4,6], 'Fl': [4,6], 'Mc': [4,6], 'Lv': [4,6],
    # Alkali/earth metals
    'Li': [4,6,8], 'Be': [4,6,8], 
    'Na': [4,6,8], 'Mg': [4,6,8], 
    'K': [4,6,8], 'Ca': [4,6,8], 
    'Rb': [4,6,8], 'Sr': [4,6,8], 
    'Cs': [4,6,8], 'Ba': [4,6,8], 
    'Fr': [4,6,8], 'Ra': [4,6,8]
    }

# GFN2-xTB single atom energies see io_xtb_calc.calc_xtb_ref_dict for reference.
xtb_single_atom_ref_es = {'H': -10.707211383396714, 
 'He': -47.432891698445495,
 'Li': -4.900000175455953,
 'Be': -15.486162554518229,
 'B': -25.917120371751917,
 'C': -48.847445262804705,
 'N': -71.00681805517411,
 'O': -102.57117256025786,
 'F': -125.69864294466228,
 'Ne': -161.42379378015602,
 'Na': -4.5469341628136,
 'Mg': -12.67981645403045,
 'Al': -24.63524632585141,
 'Si': -42.76062738849161,
 'P': -64.70342656532247,
 'S': -85.66881795502964,
 'Cl': -121.97572181135438,
 'Ar': -116.4386981693596,
 'K': -4.510348161503552,
 'Ca': -10.113012362120033,
 'Sc': -23.243511328068923,
 'Ti': -37.19952407328606,
 'V': -46.75256780519609,
 'Cr': -47.551829431041725,
 'Mn': -71.27032275557823,
 'Fe': -80.14401872919589,
 'Co': -95.42461229338535,
 'Ni': -127.03165993183204,
 'Cu': -101.98844165193441,
 'Zn': -14.354588513999577,
 'Ga': -29.418551497128806,
 'Ge': -49.249990531246375,
 'As': -60.93788396017277,
 'Se': -84.9113939279083,
 'Br': -110.16092538829801,
 'Kr': -116.24311016235609,
 'Rb': -4.353793155897734,
 'Sr': -12.583384450577476,
 'Y': -32.513603426171194,
 'Zr': -35.664705344038914,
 'Nb': -48.468943633382366,
 'Mo': -47.103307394758325,
 'Tc': -67.47438866812136,
 'Ru': -77.53081994875949,
 'Rh': -106.01053867238016,
 'Pd': -119.99800276414216,
 'Ag': -103.99479372376777,
 'Cd': -14.504682519374043,
 'In': -30.63832754080574,
 'Sn': -54.773706848758266,
 'Sb': -58.891664996211276,
 'Te': -81.88153581941742,
 'I': -102.84897812647684,
 'Xe': -105.67782578404142,
 'Cs': -4.04170614472273,
 'Ba': -11.800000422526582,
 'La': -32.78408804872581,
 'Ce': -24.476883526857367,
 'Pr': -24.30981661010049,
 'Nd': -24.142748696083178,
 'Pm': -23.975681784199946,
 'Sm': -23.80861387345064,
 'Eu': -23.641546957707547,
 'Gd': -23.474481043951293,
 'Tb': -23.30741258426,
 'Dy': -23.140346671193583,
 'Ho': -22.973279758167216,
 'Er': -22.80621184515164,
 'Tm': -22.63914493213882,
 'Yb': -22.472076019126465,
 'Lu': -22.305010106114292,
 'Hf': -35.70130487846373,
 'Ta': -51.83261682874072,
 'W': -60.25302489724785,
 'Re': -81.80817630252322,
 'Os': -81.31693407962229,
 'Ir': -17.446427065227073,
 'Pt': -120.7493981742078,
 'Au': -103.47454570514782,
 'Hg': -23.076132826294845,
 'Tl': -39.14861684553654,
 'Pb': -59.99677903578255,
 'Bi': -61.67878109601047,
 'Po': -74.41720155213928,
 'At': -81.64893636735967,
 'Rn': -104.97843975899828}
    
functional_groups_dict = {
'methyl':'C',
'ethyl':'CC',
'phenyl':'c1ccccc1',
'bromo':'Br',
'iodo':'I',
'chloro':'Cl',
'amino':'N',
'hydroxyl':'O',
'thiol':'S',
'carbonyl':'C#[O+]',
'cyano':'C#N',
'fluoro':'F',
'trichloro':'C(Cl)(Cl)Cl',
'trifluro':'C(F)(F)F',
'tribromo':'C(Br)(Br)Br',
'ether':'OC',
'carboxyl':'C(=O)O',
'carboxylate':'C(=O)[O-]',
'ester':'C(=O)OC',
'ketone':'C(=O)C',
'aldehyde':'C(=O)',
'amide':'C(=O)N(C)C',
'cyanimide':'N(C)N',
'phosphonate':'[PH+]([O-])(O)O',
'2-hydroxypyradine':'c1nc(ccc1)O',
'2-methylbenzoic_acid':'c1cc(ccc1C(=O)O)C'
}

solvents_dict = {
'acetone':'CC(=O)C',
'acetonitrile':'CC#N',
'aniline':'C1=CC=C(C=C1)N',
'benzaldehyde':'C1=CC=C(C=C1)C=O',
'benzene':'C1=CC=CC=C1 ',
'CH2Cl2':'C(Cl)Cl',
'CHCl3':'C(Cl)(Cl)Cl',
'CS2':'S=C=S',
'dioxane':'O1CCOCC1',
'dmf':'CN(C)C=O',
'dmso':'CS(=O)C',
'ether':'CCOCC',
'ethylacetate':'CCOC(=O)C',
'furane':'C1=COC=C1',
'hexandecane':'CCCCCCCCCCCCCCCC',
'hexane':'CCCCCC',
'methanol':'CO',
'nitromethane':'C[N+](=O)[O-]',
'octanol':'CCCCCCCCO', 
'phenol':'Oc1ccccc1',
'toluene':'CC1=CC=CC=C1',
'thf':'C1CCCO1',
'water':'O'
}

# Commonly used ligands dictionary largely from : https://en.wikipedia.org/wiki/Ligand (accessed 5/24/2022)
ligands_dict = {
'water':{'smiles':'O','coordList':[0],'ligType':'mono'},
'hydroxyl':{'smiles':'[OH-]','coordList':[0],'ligType':'mono'},
'oxo':{'smiles':'[O-2]','coordList':[0],'ligType':'mono'},
'hydride':{'smiles':'[H-]','coordList':[0],'ligType':'mono'},
'sulfide':{'smiles':'[S-2]','coordList':[0],'ligType':'mono'},
'pyradine':{'smiles':'c1ccncc1','coordList':[3],'ligType':'mono'},
'bipyradine':{'smiles':'n1ccccc1-c2ccccn2','coordList':[0,11],'ligType':'bi_cis'},
'bipy':{'smiles':'n1ccccc1-c2ccccn2','coordList':[0,11],'ligType':'bi_cis'},
'terpy':{'smiles':'c1ccnc(c1)c2cccc(n2)c3ccccn3','coordList':[3,11,17],'ligType':'tri_mer'},
'terpyradine':{'smiles':'c1ccnc(c1)c2cccc(n2)c3ccccn3','coordList':[3,11,17],'ligType':'tri_mer'},
'fluoride':{'smiles':'[F-]','coordList':[0],'ligType':'mono'},
'chloride':{'smiles':'[Cl-]','coordList':[0],'ligType':'mono'},
'bromide':{'smiles':'[Br-]','coordList':[0],'ligType':'mono'},
'iodide':{'smiles':'[I-]','coordList':[0],'ligType':'mono'},
'acac':{'smiles':'CC(=CC(=O)C)[O-]','coordList':[4,6],'ligType':'bi_cis'},
'methanol':{'smiles':'CO','coordList':[1],'ligType':'mono'},
'methoxy':{'smiles':'C[O-]','coordList':[1],'ligType':'mono'},
'ethanol':{'smiles':'CCO','coordList':[2],'ligType':'mono'},
'ethoxy':{'smiles':'CC[O-]','coordList':[2],'ligType':'mono'},
'octanol':{'smiles':'CCCCCCCCO','coordList':[8],'ligType':'mono'},
'nitrate_bi':{'smiles':'[N+](=O)([O-])[O-]','coordList':[2,3],'ligType':'bi_cis_chelating'},
'nitrate_mono':{'smiles':'[N+](=O)([O-])[O-]','coordList':[2],'ligType':'mono'},
'thiocyanite':{'smiles':'[S-]C#N','coordList':[0],'ligType':'mono'},
'isothiocyanite':{'smiles':'S=C=[N-]','coordList':[2],'ligType':'mono'},
'azide':{'smiles':'[N-]=[N+]=[N-]','coordList':[0],'ligType':'mono'},
'benzoate':{'smiles':'C1=CC=C(C=C1)C(=O)[O-]','coordList':[7,8],'ligType':'bi_cis_chelating'}, 
'benzoate_mono':{'smiles':'C1=CC=C(C=C1)C(=O)[O-]','coordList':[8],'ligType':'mono'},
'benzoic_acid':{'smiles':'C1=CC=C(C=C1)C(=O)O','coordList':[7],'ligType':'mono'},
'oxalate':{'smiles':'C(=O)(C(=O)[O-])[O-]','coordList':[4,5],'ligType':'bi_cis'},
'nitrite_o':{'smiles':'N(=O)[O-]','coordList':[2],'ligType':'mono'},
'nitrite_n':{'smiles':'N(=O)[O-]','coordList':[0],'ligType':'mono'},
'acetonitrile':{'smiles':'CC#N','coordList':[2],'ligType':'mono'},
'ammonia':{'smiles':'N','coordList':[0],'ligType':'mono'},
'thf':{'smiles':'C1CCCO1','coordList':[4],'ligType':'mono'},
'en':{'smiles':'NCCN','coordList':[0,3],'ligType':'bi_cis'},
'ethylenediamine':{'smiles':'NCCN','coordList':[0,3],'ligType':'bi_cis'},
'phen':{'smiles':'c1cc2ccc3cccnc3c2nc1','coordList':[9,12],'ligType':'bi_cis'},
'phenanthroline':{'smiles':'c1cc2ccc3cccnc3c2nc1','coordList':[9,12],'ligType':'bi_cis'},
'tpp':{'smiles':'c1ccccc1P(c2ccccc2)c3ccccc3','coordList':[6],'ligType':'mono'},
'triphenylphosphine':{'smiles':'c1ccccc1P(c2ccccc2)c3ccccc3','coordList':[6],'ligType':'mono'},
'cyanide':{'smiles':'[C-]#N','coordList':[0],'ligType':'mono'},
'isocyanide':{'smiles':'[C-]#N','coordList':[1],'ligType':'mono'},
'carbonyl':{'smiles':'[C-]#[O+]','coordList':[0],'ligType':'mono'},
'dppe':{'smiles':'P(c1ccccc1)(c2ccccc2)CCP(c3ccccc3)c4ccccc4','coordList':[0,15],'ligType':'bi_cis'},
'bisdiphenylphosphinoethane':{'smiles':'P(c1ccccc1)(c2ccccc2)CCP(c3ccccc3)c4ccccc4','coordList':[0,15],'ligType':'bi_cis'},
'dppm':{'smiles':'P(c1ccccc1)(c2ccccc2)CP(c3ccccc3)c4ccccc4','coordList':[0,14],'ligType':'bi_cis'},
'bisdiphenylphosphinomethane':{'smiles':'P(c1ccccc1)(c2ccccc2)CP(c3ccccc3)c4ccccc4','coordList':[0,14],'ligType':'bi_cis'},
'corrole':{'smiles':'C1=CC2=CC3=CC=C([N-]3)C4=NC(=CC5=CC=C([N-]5)C=C1[N-]2)C=C4','coordList':[8,10,17,20],'ligType':'tetra_planar'},
'9-crown-3':{'smiles':'C1COCCOCCO1','coordList':[2,5,8],'ligType':'tri_fac'},
'12-crown-4':{'smiles':'O1CCOCCOCCOCC1', 'coordList':[0,3,6,9],'ligType':'tetra_pyramidal'},
'15-crown-5':{'smiles':'C1COCCOCCOCCOCCO1','coordList':[2,5,8,11,14],'ligType':'penta_planar'},
'18-crown-6':{'smiles':'O1CCOCCOCCOCCOCCOCC1','coordList':[0,3,6,9,12,15],'ligType':'hexa_planar'},
'222-cryptand':{'smiles':'C1COCCN2CCOCCOCCN(CCO1)CCOCCOCC2','coordList':[2,5,8,11,14,17,20,23],
              'ligType':'octa_trigonal_prismatic_triangle_face_bicapped'},
'cp':{'smiles':'C1=C[CH-]C=C1','coordList':[0,1,2,3,4],'ligType':'sandwich'},
'cyclopentadienyl':{'smiles':'C1=C[CH-]C=C1','coordList':[0,1,2,3,4],'ligType':'sandwich'},
'benzene':{'smiles':'c1ccccc1','coordList':[0,1,2,3,4,5],'ligType':'sandwich'},
'cp_m':{'smiles':'C[C-]1C(=C(C(=C1C)C)C)C','coordList':[1,2,3,4,5],'ligType':'sandwich'},
'pentamethylcyclopentadienyl':{'smiles':'C[C-]1C(=C(C(=C1C)C)C)C','coordList':[1,2,3,4,5],'ligType':'sandwich'},
'dien':{'smiles':'NCCNCCN','coordList':[0,3,6],'ligType':'tri_fac'},
'diethylenetriamine':{'smiles':'NCCNCCN','coordList':[0,3,6],'ligType':'tri_fac'},
'dmgh-':{'smiles':'C\C(=N\O)\C(\C)=N\[O-]','coordList':[2,6],'ligType':'bi_cis'},
'dota':{'smiles':'O=C(O)CN(CC1)CCN(CC(=O)O)CCN(CC(=O)O)CCN1CC(=O)O','coordList':[0,4,9,12,16,23,19,26],
        'ligType':'octa_square_antiprismatic'}, 
'tetraxetan':{'smiles':'O=C(O)CN(CC1)CCN(CC(=O)O)CCN(CC(=O)O)CCN1CC(=O)O','coordList':[0,4,9,12,16,23,19,26],
        'ligType':'octa_square_antiprismatic'},
'dtpa':{'smiles':'C(CN(CC(=O)O)CC(=O)O)N(CCN(CC(=O)O)CC(=O)O)CC(=O)O', 
        'coordList':[2,5,9,11,14,17,21,25],'ligType':'octa_square_antiprismatic'},
'3,4,3-li(1-2-hopo)':{'smiles':'C1=CC(=O)N(C(=C1)C(=O)NCCCN(CCCCN(CCCNC(=O)C2=CC=CC(=O)N2[O-])C(=O)C3=CC=CC(=O)N3[O-])C(=O)C4=CC=CC(=O)N4[O-])[O-]', 
        'coordList':[3,53,52,50,30,32,40,42],'ligType':'octa_square_antiprismatic'},
'dheba':{'smiles':'CCCCC(CC)COP(=O)([O-])OCC(CC)CCCC','coordList':[10,11],'ligType':'bi_cis_bulky'},
'edta':{'smiles':'[O-]C(=O)CN(CCN(CC([O-])=O)CC([O-])=O)CC([O-])=O','coordList':[0,4,7,10,14,18],
        'ligType':'hexa_octahedral'},
'ethylenediaminetriacetate':{'smiles':'C(C[NH+](CC(=O)[O-])CC(=O)[O-])NCC(=O)[O-]',
        'coordList':[2,6,10,11,15],'ligType':'penta_square_pyramidal'},
'egta':{'smiles':'O=C([O-])CN(CC(=O)[O-])CCOCCOCCN(CC(=O)[O-])CC(=O)[O-]',
        'coordList':[2,4,8,11,14,17,21,25],'ligType':'octa_trigonal_prismatic_triangle_face_bicapped'},
'glycine':{'smiles':'C(C(=O)[O-])N','coordList':[3,4],'ligType':'bi_cis'},
'porphyrin':{'smiles':'C1=CC2=CC3=CC=C([N-]3)C=C4C=CC(=N4)C=C5C=CC(=N5)C=C1[N-]2','coordList':[8,14,20,23],
             'ligType':'tetra_planar'},
'corrin':{'smiles':'C1CC2=NC1C3CCC(=N3)C=C4CCC(=N4)C=C5CCC(=C2)[N-]5','coordList':[3,9,15,22],
          'ligType':'tetra_planar'},
'ida':{'smiles':'[O-]C(=O)CNCC([O-])=O','coordList':[0,4,7],'ligType':'tri_mer'},
'iminodiacetate':{'smiles':'[O-]C(=O)CNCC([O-])=O','coordList':[0,4,7],'ligType':'tri_mer'},
'nitrosyl':{'smiles':'[N+]=O','coordList':[0],'ligType':'mono'},
'nta':{'smiles':'O=C([O-])CN(CC(=O)[O-])CC(=O)[O-]','coordList':[2,4,7,12],'ligType':'tetra_seesaw'},
'pyrazine':{'smiles':'c1cnccn1','coordList':[2],'ligType':'mono'},
'scorpionate':{'smiles':'[CH-]1C=C[N+](=C(C(=O)[O-])N2C=CC=N2)[N-]1','coordList':[7,12,13],
               'ligType':'tri_fac'},
'tp':{'smiles':'[BH-](N1C=CC=N1)(N2C=CC=N2)N3C=CC=N3','coordList':[5,10,15],'ligType':'tri_fac'},
'trithia-9-crown-3':{'smiles':'C1CSCCSCCS1','coordList':[2,5,8],'ligType':'tri_fac'},
'sulfite':{'smiles':'[O-]S(=O)[O-]','coordList':[1],'ligType':'mono'},
'isosulfite':{'smiles':'[O-]S(=O)[O-]','coordList':[0],'ligType':'mono'},
'tacn':{'smiles':'C1CNCCNCCN1','coordList':[2,5,8],'ligType':'tri_fac'},
'tricyclhexylphosphine':{'smiles':'C1(CCCCC1)P(C2CCCCC2)C3CCCCC3','coordList':[6],'ligType':'mono'},
'trien':{'smiles':'NCCNCCNCCN','coordList':[0,3,6,9], 'ligType':'tetra_trigonal_pyramidal'},
'tmp':{'smiles':'CP(C)C','coordList':[1],'ligType':'mono'},
'triso-tolylphosphine':{'smiles':'CC1=CC=CC=C1P(C2=CC=CC=C2C)C3=CC=CC=C3C','coordList':[7],'ligType':'mono'},
'tren':{'smiles':'NCCN(CCN)CCN','coordList':[0,3,6,9],'ligType':'tetra_trigonal_pyramidal'},
'tris2-diphenylphosphinoethylamine':{'smiles':'c1ccc(cc1)P(CCN(CCP(c2ccccc2)c3ccccc3)CCP(c4ccccc4)c5ccccc5)c6ccccc6',
                                     'coordList':[6,9,12,27], 'ligType':'tetra_trigonal_pyramidal'},
# 'trpoylium':{'smiles':'c1=cc=c[cH+]c=c1','coordList':[0]} # Can't get smiles to work with obabel/mmff94
'amine':{'smiles':'N','coordList':[0],'ligType':'mono'},
'co2':{'smiles':'O=C=O','coordList':[1],'ligType':'mono'},
'phosphorus_trifluoride':{'smiles':'FP(F)F','coordList':[1],'ligType':'mono'},
'topo':{'smiles':'CCCCCCCCP(=O)(CCCCCCCC)CCCCCCCC','coordList':[9],'ligType':'mono'},
'trimethyl_phosphite':{'smiles':'COP(OC)OC','coordList':[2],'ligType':'mono'},
'trimethyl_phosphine_oxide':{'smiles':'CP(=O)(C)C','coordList':[1],'ligType':'mono'}, 
'cyanex-301':{'smiles':'CC(CC(C)(C)C)CP(=S)(CC(C)CC(C)(C)C)[S-]','coordList':[9,18],'ligType':'bi_cis_chelating'},
'bmptt':{'smiles':'CC1=NN(C(=S)[C@H]1C(=O)c1ccccc1)c1ccccc1','coordList':[5,8],'ligType':'bi_cis'},
'dpphen':{'smiles':'C1=CC=C(C=C1)C2=C3C=CC4=C(C=CN=C4C3=NC=C2)C5=CC=CC=C5','coordList':[14,17],'ligType':'bi_cis'},
'bis_4-chlorophenyl_dithiophosphinate':{'smiles':'C1=CC(=CC=C1P(=S)(C2=CC=C(C=C2)Cl)[S-])Cl',
                                        'coordList':[7,15],'ligType':'bi_cis'},
# 'talspeak-hdehdga':{'smiles':'CCCCC(CC)CC(CC(CC)CCCC)(C(=O)O)OCC(=O)N'} # Multiprotic
'btp':{'smiles':'c1cnnc(n1)c1cccc(n1)c1nccnn1','coordList':[3,11,17],'ligType':'tri_mer'},
'kryptofix-22':{'smiles':'C1COCCOCCN(CCOCCOCCN1C)C','coordList':[2,5,8,11,14,17],'ligType':'hexa_planar'}, # Functionalize at 18,19
'kryptofix-222':{'smiles':'C1COCCOCCN2CCOCCOCCN1CCOCCOCC2','coordList':[2,5,8,11,14,17,20,23], #
                    'ligType':'octa_trigonal_prismatic_triangle_face_bicapped'},
'kryptofix-211':{'smiles':'C1COCCN2CCOCCN1CCOCCOCC2','coordList':[2,5,8,11,14,17], #
                    'ligType':'hexa_trigonal_prismatic'},
'oxalacetic_acid':{'smiles':'C(C(=O)C(=O)O)C(=O)O','coordList':[2,4,7],'ligType':'tri_fac'}, #
'oxalacetic_acid_bi':{'smiles':'C(C(=O)C(=O)O)C(=O)O','coordList':[2,4],'ligType':'bi_cis'},
'oxalacetate':{'smiles':'C(C(=O)C(=O)[O-])C(=O)O','coordList':[2,5,7],'ligType':'tri_fac'}, #
'oxalacetate_bi':{'smiles':'C(C(=O)C(=O)[O-])C(=O)O','coordList':[2,5,7],'ligType':'bi_cis'},
'oxalbiacetate':{'smiles':'C(C(=O)C(=O)[O-])C(=O)[O-]','coordList':[2,5,8],'ligType':'tri_fac'},
'oxalbiacetate_bi':{'smiles':'C(C(=O)C(=O)[O-])C(=O)[O-]','coordList':[2,5],'ligType':'bi_cis'},
'valine':{'smiles':'CC(C)C(C(=O)O)N','coordList':[5,6],'ligType':'bi_cis'}, #
'valinate_bi_n':{'smiles':'CC(C)C(C(=O)[O-])N','coordList':[6,7],'ligType':'bi_cis'},
'valinate_bi_o':{'smiles':'CC(C)C(C(=O)[O-])N','coordList':[5,7],'ligType':'bi_cis_chelating'},
'oxime':{'smiles':'c1(O)c(C(C)(C)C)cc(C(C)(C)C)cc1(C(=NO))','coordList':[1,16],'ligType':'bi_cis'},
'pyridine_ester':{'smiles':'COC(=O)C1=CC(=CN=C1)C(=O)OC','coordList':[8],'ligType':'mono'},
'tedga':{'smiles':'CCN(CC)C(=O)COCC(=O)N(CC)CC','coordList':[6,8,11],'ligType':'tri_mer_bent'},
'hydroxyquinoline':{'smiles':'C1=CC2=C(C(=C(C)1)O)N=CC=C2','coordList':[7,8],'ligType':'bi_cis'}, # Functionalize 6
'bi_benzimidazole':{'smiles':'C1=CC=C2C(=C1)NC(=N2)C3=NC4=CC=CC=C4N3','coordList':[8,10],'ligType':'bi_cis'} # Functionalize 6,17 with esters
}


def convert_actinides_lanthanides(elem,inverse=False):
    """convert_actinides_lanthanides [summary]

    Parameters
    ----------
    elem : str
        element (metal) to consider
    inverse : bool, optional
        wheter to re-convert lanthanides to actinides in final structure, by default False

    Returns
    -------
    out : str
        new metal symbol
    isact : bool
        whether the elem passed was an actinide or is an actinide complex
    """
    acts = actinides
    lns = lanthanides
    isact = False
    if elem in acts:
        out = lns[acts.index(elem)]
        isact = True
    elif inverse:
        out = acts[lns.index(elem)]
        isact = True
    else:
        out = elem
    return out, isact

def larger_map_metal(inputDict):
    """larger_map_metal 
    Map metal to "similar" metal with covalent radii about 1.3 times the normal one

    Parameters
    ----------
    inputDict : dict
        input dictionary
    
    Returns
    ----------
    newinpDict : dict
        new input dictionary
    match : bool
        whether or not a larger element exists to try
    metal : str 
        what the original string was
    """
    metal = inputDict['core']['smiles'].strip('[').strip(']')
    metal, isact = convert_actinides_lanthanides(metal) # Ensure lanthanide used as reference for size not actinide
    covrad_metal = rcov1[elements.index(metal)]
    if metal in transition_metals:
        rcov_tms = np.array([limited_rcov1[limited_elements.index(x)] for x in limited_transition_metals])
        matches = np.isclose(rcov_tms,1.3*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_transition_metals[np.where(matches)[0][0]]
        else:
            match = None
    elif metal in post_transition_metals:
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_post_transition_metals]
        matches = np.isclose(rcov_tms,1.3*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_post_transition_metals[np.where(matches)[0][0]]
        else:
            match = None
    elif metal in alkali_and_alkaline_earth:
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_alkali_and_alkaline_earth]
        matches = np.isclose(rcov_tms,1.3*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_alkali_and_alkaline_earth[np.where(matches)[0][0]]
        else:
            match = None
    elif metal in heavy_metals:
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_heavy_metals]
        matches = np.isclose(rcov_tms,1.3*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_heavy_metals[np.where(matches)[0][0]]
        else:
            match = None
    if (not match): # Try to see if any metals within tolerance
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_all_metals]
        matches = np.isclose(rcov_tms,1.3*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_all_metals[np.where(matches)[0][0]]
        else:
            matches = np.isclose(rcov_tms,1.3*covrad_metal,atol=1e-1)
            if np.any(matches): # Try again with looser tolerance
                match = limited_all_metals[np.where(matches)[0][0]]
            else:
                match = None
    if isinstance(match,str):
        newinpDict = copy.deepcopy(inputDict)
        newinpDict['core']['smiles'] = '['+match+']'
        newinpDict['core']['metal'] = match
        match = True
    else:
        match = False
        newinpDict = copy.deepcopy(inputDict)
    return newinpDict, match, metal

def smaller_map_metal(inputDict):
    """larger_map_metal 
    Map metal to "similar" metal with covalent radii about 0.8 times the normal one

    Parameters
    ----------
    inputDict : dict
        input dictionary
    
    Returns
    ----------
    newinpDict : dict
        new input dictionary
    match : bool
        whether or not a larger element exists to try
    metal : str 
        what the original string was
    """
    metal = inputDict['core']['smiles'].strip('[').strip(']')
    metal, isact = convert_actinides_lanthanides(metal)
    covrad_metal = rcov1[elements.index(metal)]
    if metal in transition_metals:
        rcov_tms = np.array([limited_rcov1[limited_elements.index(x)] for x in limited_transition_metals])
        matches = np.isclose(rcov_tms,0.8*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_transition_metals[np.where(matches)[0][0]]
        else:
            match = None
    elif metal in post_transition_metals:
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_post_transition_metals]
        matches = np.isclose(rcov_tms,0.8*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_post_transition_metals[np.where(matches)[0][0]]
        else:
            match = None
    elif metal in alkali_and_alkaline_earth:
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_alkali_and_alkaline_earth]
        matches = np.isclose(rcov_tms,0.8*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_alkali_and_alkaline_earth[np.where(matches)[0][0]]
        else:
            match = None
    elif metal in heavy_metals:
        rcov_tms = [limited_rcov1[elements.index(x)] for x in limited_heavy_metals]
        matches = np.isclose(rcov_tms,0.8*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_heavy_metals[np.where(matches)[0][0]]
        else:
            match = None
    if (not match): # Try to see if any metals within tolerance
        rcov_tms = [limited_rcov1[limited_elements.index(x)] for x in limited_all_metals]
        matches = np.isclose(rcov_tms,0.8*covrad_metal,atol=5e-2)
        if np.any(matches):
            match = limited_all_metals[np.where(matches)[0][0]]
        else:
            matches = np.isclose(rcov_tms,0.8*covrad_metal,atol=1e-1)
            if np.any(matches): # Try again with looser tolerance
                match = limited_all_metals[np.where(matches)[0][0]]
            else:
                match = None
    if isinstance(match,str):
        newinpDict = copy.deepcopy(inputDict)
        newinpDict['core']['smiles'] = '['+match+']'
        newinpDict['core']['metal'] = match
        match = True
    else:
        match = False
        newinpDict = copy.deepcopy(inputDict)
    return newinpDict, match, metal

def map_metal_radii(inputDict,larger=False,
                    larger_factor=1.3,smaller_factor=0.8):
    """map_metal_radii
    Rescale metal covalent radii and vdwrad to about 1.3 times the normal one

    Parameters
    ----------
    inputDict : dict
        input dictionary
    larger : bool
        wheter to make the radii larger or smaller, default is False
    larger_factor : float
        scale by which to make the radii larger, default is 1.3
    smaller_factor : float
        scale by which to make the radii smaler, default is 0.8

    
    Returns
    ----------
    newinpDict : dict
        new input dictionary
    match : bool
        whether or not a larger element exists to try
    metal : str 
        what the original string was 
    """
    metal = inputDict['core']['smiles'].strip('[').strip(']')
    covrad_metal = rcov1[elements.index(metal)]
    vdwrad_metal = rvdw[elements.index(metal)]
    newinpDict = copy.deepcopy(inputDict)
    if larger:
        newinpDict['parameters']['covrad_metal'] = covrad_metal*larger_factor
        newinpDict['parameters']['rvdw_metal'] = vdwrad_metal*larger_factor
        newinpDict['parameters']['scaled_radii_factor'] = larger_factor
    else:
        newinpDict['parameters']['covrad_metal'] = covrad_metal*smaller_factor
        newinpDict['parameters']['rvdw_metal'] = vdwrad_metal*smaller_factor
        newinpDict['parameters']['scaled_radii_factor'] = smaller_factor
    return newinpDict


from ase.vibrations import Vibrations
from ase.thermochemistry import IdealGasThermo
import ase
import numpy as np
from scipy.stats import bernoulli
from architector import arch_context_manage


def calc_free_energy(relaxed_atoms ,temp=298.15, pressure=101325, geometry='nonlinear'):
    """calc_free_energy utility function to calculate free energy of relaxed structures with
    ASE calculators added.

    Uses the ideal gas rigid rotor harmonic oscillator (IGRRHO) approach

    Parameters
    ----------
    relaxed_atoms : ase.atoms.Atoms
        relaxed structures with calculator attached (usually XTB)
    temp : float, optional
        temperature in kelvin, by default 298.15
    pressure : int, optional
        pressure in pascal, by default 101325 Pa
    geometry : str, optional
        'linear','nonlinear' , by default 'nonlinear'

    Returns
    -------
    G, float
       free energy in eV
    thermo, ase.thermo
        ASE thermo calculator
    """
    with arch_context_manage.make_temp_directory() as _:
        potentialenergy = relaxed_atoms.get_potential_energy()
        vib = Vibrations(relaxed_atoms)
        vib.run()
        vib_energies = vib.get_energies()
        nunpaired = np.sum(relaxed_atoms.get_initial_magnetic_moments())

        thermo = IdealGasThermo(vib_energies=vib_energies,
                                potentialenergy=potentialenergy,
                                atoms=relaxed_atoms,
                                geometry=geometry,
                                symmetrynumber=2, spin=nunpaired/2)
        G = thermo.get_gibbs_energy(temperature=temp, pressure=pressure)
    return G, thermo


def normal_mode_sample(relaxed_atoms,temp=298.15,n=10,seed=42): # T=1 produced okay results in preliminary tests
    """normal_mode_sample
    https://www.nature.com/articles/sdata2017193#Sec12

    Parameters
    ----------
    relaxed_atoms : ase.atoms.Atoms
        relaxed atoms with calculator attached
    temp : int, optional
        temperature for kbT approximation, by default 298.15, possibly lower later
    n : int, optional
        number of conformations to generate, by default 10
    seed : int, optional
        random seed for generation, by default 42

    Returns
    -------
    displaced_structures : list (ase.atoms.Atoms)
        A list of the displaced structures that have been evaluated by the calculator

    Raises
    ------
    ValueError
        If the imaginary modes are quite strong
    """
    good = True
    with arch_context_manage.make_temp_directory() as _:
        vib = Vibrations(relaxed_atoms)
        calc = relaxed_atoms.get_calculator()
        vib.run()
        vib_energies = vib.get_energies()
        if np.any(np.imag(vib_energies) > 0.1):
            print('Warning: There are some highly imaginary modes!')
            good = False
        vib_energies = np.real(vib_energies)
        non_zero_inds = np.where(np.real(vib_energies) > 0.0001)[0]
        kbT = ase.units.kB*temp
        # Seed
        if seed:
            np.random.seed(seed)
        na = len(relaxed_atoms)
        displaced_structures = []
        for i in range(n):
            out_atoms = relaxed_atoms.copy()
            # Generate random cs
            cs = np.random.random(len(non_zero_inds))
            scale = np.random.random()
            cs = cs/np.sum(cs)*scale # Set to scale between 0-1
            # Generate random signs
            signs = bernoulli.rvs(0.5,size=len(non_zero_inds))
            signs[np.where(signs == 0)] = -1
            Rs = signs*np.sqrt(3*cs*na*kbT/vib_energies[non_zero_inds])
            displaced_posits = relaxed_atoms.get_positions()
            for j,r in enumerate(Rs):
                norm_mode = vib.get_mode(non_zero_inds[j])
                displacement = norm_mode*r
                displaced_posits += displacement
            out_atoms.set_positions(displaced_posits)
            out_atoms.calc = calc
            try:
                out_atoms.get_total_energy()
                displaced_structures.append(out_atoms)
            except:
                continue
    if good:
        return displaced_structures
    else:
        return []


"""
IO module for CREST.

Developed by Dan Burrill and Michael Taylor
"""

# Imports
import ase.io as ase_io
import shutil
import copy
import subprocess as sub
import numpy as np
from architector import io_ptable

import architector.io_obabel as io_obabel 
import architector.io_molecule as io_molecule
import architector.arch_context_manage as arch_context_manage
import architector.io_xtb_calc as io_xtb_calc
from ase.optimize import BFGSLineSearch

def isint(str):
    out = False
    try:
        out = int(str)
        out = True
    except:
        out = False
    return out

# Functions
def read_conformers(fileName):
    '''
    Read conformers from file.

    Parameters
    ----------
    fileName : str
        File path to read in conformers generated by crest
    
    Returns
    ----------
    molList : list (ase.atoms.Atoms)
        Conformers read from crest as ASE atoms
    xtb_energies : list (float)
        Sorted xtb energies from the output
    '''
    # Read conformers to ASE Atoms objects
    molList = ase_io.read(fileName,index=":")

    # Center all conformer COMs at the origin
    for mol in molList:
        mol.translate(-mol.get_center_of_mass())

    with open(fileName,'r') as file1:
        lines = file1.readlines()
    xtb_energies = []
    read = False
    for line in lines:
        sline = line.split()
        if len(sline) == 1:
            is_int = isint(sline[0])
            if is_int:
                if int(sline[0]) == len(mol):
                    read=True
            elif read:
                xtb_energies.append(float(sline[0])*27.2114) # Hartee to eV
                read = False
            else:
                print('Warning - messed up!')
    return molList,xtb_energies

def crest_conformers(structure, charge=None, uhf=None, method='GFN2/GFNFF',
                    solvent='none'):
    '''
    Find conformers of a given structure with CREST.

    Parameters
    ----------
    structure : ase.atoms.Atoms/str, optional
        structure passsed, default None
    charge : int, optional
        charge of the species, default None
    uhf : int, optional
        number of unpaired electrons in the system, default None
    method : str, optional
        which gfn family method to use, default GFN2/GFNFF
    solvent : str, optional
        whether to use a solvent for conformer evalulation, default 'none'
    read_charge_spin : bool, optional
        whether to read in the charge/spin from the structure (mol2), default False

    Returns
    ----------
    conformerList : list (ase.atoms.Atoms)
        Conformers generated via crest as ASE atoms
    xtb_energies : list (float)
        List of xtb energies output from CREST
    '''

    # Convert smiles to xyz string

    crestPath = shutil.which('crest')

    mol = io_molecule.convert_io_molecule(structure)
    if charge is not None:
        mol.charge = charge
    elif mol.xtb_charge is None:
        mol.detect_charge_spin()
    mol_charge = mol.xtb_charge

    even_odd_electrons = (np.sum([atom.number for atom in mol.ase_atoms])-mol_charge) % 2
    if (uhf is not None):
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1
    elif mol.xtb_uhf is not None:
        uhf = mol.xtb_uhf
    else:
        uhf = 0 # Set spin to LS by default
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1

    mol_charge = int(mol_charge) # Ensure integers
    uhf = int(uhf)

    xyzstr = io_molecule.convert_ase_xyz(mol.ase_atoms)

    with arch_context_manage.make_temp_directory() as _:
        # Write xyz file
        with open("structure.xyz",'w') as outFile:
            outFile.write(xyzstr)

        if method == 'GFN2/GFNFF':
            method_string = '--gfn2//gfnff'
        elif method == 'GFN2-xTB':
            method_string = '--gnf2'
        elif method == 'GFN-FF':
            method_string = '--gfnff'
        else:
            raise ValueError("This crest method is not recognized {}.".format(method))

        # Run CREST
        if (uhf == 0) and (solvent == 'none'):
            execStr = "{} structure.xyz {} --chrg {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge)
        elif (uhf == 0) and (solvent != 'none'):
            execStr = "{} structure.xyz {} --chrg {} --alpb {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge,solvent)
        elif (solvent == 'none'):
            execStr = "{} structure.xyz {} --chrg {} --uhf {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge,uhf)
        else:
            execStr = "{} structure.xyz {} --chrg {} --uhf {} --alpb {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge,uhf,solvent)

        sub.run(execStr,shell=True,check=True)

        # Read conformers from file
        conformerList, xtb_energies = read_conformers("crest_conformers.xyz")

    return conformerList, xtb_energies

def crest_conformers_lig(smiles,ase_atoms=None,charge=None,uhf=0,method='GFN2/GFNFF',
                    solvent='none',neutralize=False,functionalizations=None):
    '''
    Find conformers of a given smiles with CREST.

    Parameters
    ----------
    smiles : str
        smiles string for ligand
    ase_atoms : ase.atoms.Atoms, optional
        structure passsed, default None
    charge : int, optional
        charge of the species, default to initial charges set on smiles
    uhf : int, optional
        number of unpaired electrons in the system, default to 0
    method : str, optional
        which gfn family method to use, default GFN2/GFNFF
    solvent : str, optional
        whether to use a solvent for conformer evalulation, default 'none'
    neutralize: bool, optional
        neutralize ligand before evaluating conformers? default False.
    functionalizations: bool, optional
        add functionalizations to the ligand, default False

    Returns
    ----------
    conformerList : list (ase.atoms.Atoms)
        Conformers generated via crest as ASE atoms
    xtb_energies : list (float)
        List of xtb energies output from CREST
    '''

    # Convert smiles to xyz string

    crestPath = shutil.which('crest')

    if ase_atoms is None:
        struct = io_obabel.get_obmol_smiles(
            smiles,
            neutralize=neutralize,
            functionalizations=functionalizations
            )
        if charge:
            mol_charge = charge
        else:
            mol_charge = struct.GetTotalCharge()
        ase_atoms = io_obabel.convert_obmol_ase(struct)
    else: 
        mol_charge = np.sum(ase_atoms.get_initial_charges())

    even_odd_electrons = (np.sum([atom.number for atom in ase_atoms])-mol_charge) % 2
    if (uhf is not None):
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1
    else:
        uhf = 0 # Set spin to LS by default
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1

    xyzstr = io_molecule.convert_ase_xyz(ase_atoms)
    mol_charge = int(mol_charge) # Ensure integers
    uhf = int(uhf)

    with arch_context_manage.make_temp_directory() as _:
        # Write xyz file
        with open("structure.xyz",'w') as outFile:
            outFile.write(xyzstr)

        if method == 'GFN2/GFNFF':
            method_string = '--gfn2//gfnff'
        elif method == 'GFN2-xTB':
            method_string = '--gnf2'
        elif method == 'GFN-FF':
            method_string = '--gfnff'
        else:
            raise ValueError("This crest method is not recognized {}.".format(method))

        # Run CREST

        if (uhf == 0) and (solvent == 'none'):
            execStr = "{} structure.xyz {} --chrg {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge)
        elif (uhf == 0) and (solvent != 'none'):
            execStr = "{} structure.xyz {} --chrg {} --alpb {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge,solvent)
        elif (solvent == 'none'):
            execStr = "{} structure.xyz {} --chrg {} --uhf {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge,uhf)
        else:
            execStr = "{} structure.xyz {} --chrg {} --uhf {} --alpb {} --notopo --quick > output.crest".format(
                                                                            crestPath,method_string,mol_charge,uhf,solvent)

        sub.run(execStr,shell=True,check=True)

        # Read conformers from file
        conformerList, xtb_energies = read_conformers("crest_conformers.xyz")

    return conformerList, xtb_energies


def add_explicit_solvents(complex_mol2string,n_solvents=6,solvent='water'):
    """
    THIS IS HIGHLY EXPERIMENTAL:
    To use this xtb_IFF is needed: https://github.com/grimme-lab/xtbiff/releases/tag/v1.1 

    We were attempting this type of solvent addition protocol:
    https://xtb-docs.readthedocs.io/en/latest/crestqcg.html to add solvents!
    """
    
    if solvent in io_ptable.solvents_dict:
        solvent_smiles = io_ptable.solvents_dict[solvent]
    elif solvent in io_ptable.solvents_dict.values():
        solvent_smiles = copy.deepcopy(solvent)
        for key, value in io_ptable.solvents_dict.items():
            if value == solvent_smiles:
                solvent = key
    else:
        raise ValueError('Crest/XTB does not know this solvent!')

    mol = io_molecule.Molecule()
    charge,spin = mol.read_mol2(complex_mol2string,readstring=True,read_charge_spin=True)
    
    info_dict = dict()
    if charge is None:
        _,_,info_dict = io_obabel.obmol_lig_split(complex_mol2string,return_info=True)
        charge = int(io_ptable.metal_charge_dict[info_dict['metal']] + np.sum(info_dict['lig_charges']))
    if spin is None:
        even_odd_electrons = (np.sum([atom.number for atom in mol.ase_atoms])-charge) % 2
        if len(info_dict) > 1: # Check for info_dict
            uhf = io_ptable.metal_spin_dict[info_dict['metal']]
        else: # Otherwise calc the metal id!
             _,_,info_dict = io_obabel.obmol_lig_split(complex_mol2string,return_info=True)
             uhf = io_ptable.metal_spin_dict[info_dict['metal']]
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1
        spin = int(uhf)

    charge = int(charge)
    spin = int(spin)
    execStr = "crest solute.xyz --qcg solvent.xyz --chrg {} --uhf {} --nsolv {} --T 12 --ensemble --alpb {} --mdtime 50 --mddump 200 > crest.out".format(
        charge,spin,n_solvents,solvent)

    solvent_xyz_str = io_obabel.smiles2xyz(solvent_smiles)
    solute_xyz_str = mol.write_xyz('cool.xyz',writestring=True)

    metal_ind = [i for i,x in enumerate(mol.ase_atoms.get_chemical_symbols()) if x in io_ptable.all_metals][0]
    coord_atoms = np.nonzero(np.ravel(mol.graph[metal_ind]))[0]
    freeze_string = ','.join([str(x) for x in sorted(coord_atoms.tolist() + [metal_ind])])

    molcontrol_str = """$constrain
  atoms: {}
$end""".format(freeze_string)

    # os.mkdir('test') # For development
    # os.chdir('test')

    with arch_context_manage.make_temp_directory() as _:
        with open('solute.xyz','w') as file1:
            file1.write(solute_xyz_str)
        with open('solvent.xyz','w') as file1:
            file1.write(solvent_xyz_str)
        with open('.xcontrol','w') as file1:
            file1.write(molcontrol_str)
    
    sub.run(execStr,shell=True,check=True)

    conformerList, xtb_energies = read_conformers("final_ensemble.xyz")
        
    return conformerList, xtb_energies
        

def obmol_xtb_conformers(smiles,charge=None,uhf=0,method='GFN2-xTB',total_confs=3000,
                    solvent='none',neutralize=False,functionalizations=None):
    '''
    Relax ligand structure with xTB. (Sampling done with openbabel)

    Parameters
    ----------
    smiles : str
        smiles string for ligand
    charge : int, optional
        charge of the species, default to initial charges set on smiles
    uhf : int, optional
        number of unpaired electrons in the system, default to 0
    method : str, optional
        which gfn family method to use, default GFN2/GFNFF
    solvent : str, optional
        whether to use a solvent for conformer evalulation, default 'none'
    neutralize: bool, optional
        neutralize ligand before evaluating conformers? default False.
    functionalizations : None, optional
        functionalizations option for smiles

    Returns
    ----------
    conformerList : list (ase.atoms.Atoms)
        Conformers generated via openbabel as ASE atoms
    xtb_energies : list (float)
        List of xtb energies output 
    '''

    struct = io_obabel.get_obmol_smiles(
        smiles,
        neutralize=neutralize
        )
    conf_list = io_obabel.generate_obmol_conformers(smiles, neutralize=neutralize,
        functionalizations=functionalizations,conf_cutoff=total_confs)
    if charge:
        mol_charge = charge
    else:
        mol_charge = struct.GetTotalCharge()

    conformerList = []
    xtb_energies = []

    with arch_context_manage.make_temp_directory() as _:
        for conf in conf_list:
            mol = io_molecule.convert_io_molecule(conf)
            atoms = mol.ase_atoms
            io_xtb_calc.set_XTB_calc_lig(atoms,charge=mol_charge,method=method,uhf=uhf,
                                         solvent=solvent)
            fail = True
            try:
                dyn = BFGSLineSearch(atoms,logfile='tmp.log') # Mute output
                dyn.run()
                fail = False
            except:
                fail = True 
            if not fail:
                conformerList.append(atoms)
                xtb_energies.append(atoms.get_total_energy())

    return conformerList, xtb_energies


# Main (Unit Tests)
if (__name__ == '__main__'):
    # Variables
    smiles = "n1ccccc1-c2ccccn2"

    crestPath=shutil.which('crest')
    
    # Check conformers
    with arch_context_manage.make_temp_directory() as _:
        conformerList,energies = obmol_xtb_conformers(smiles,method='GFN2/GFNFF')

    print(conformerList)

    for idx,mol in enumerate(conformerList):
        mol.write("con_{}.xyz".format(energies[idx]))


""" 
Code containing molecule/io
Molecule class encodes graph/bond order information and can handle a variety of types of inputs.
convert_io_molecule handles conversion to molecule object.

Also contains sanity check and graph sanity check routines for checking final geometries.

Developed by Michael Taylor
"""

import ase
from ase.io import read
from ase.atoms import Atoms
from ase.atom import Atom
import numpy as np
import re
import copy
import itertools
import architector
from architector import io_obabel
import architector.io_ptable as io_ptable
from io import StringIO
from scipy.sparse.csgraph import (csgraph_from_dense, connected_components)

def convert_io_molecule(structure,
                        detect_charge_spin=False,
                        charge=None,
                        uhf=None,
                        xtb_uhf=None,
                        xtb_charge=None):
    """convert_io_molecule
    Handle multiple types of structures passed and convert them to architector.io_molecule.Molecule objects.

    Parameters
    ----------
    structure : str
        Structure to convert to mol object
    detect_charge_spin : bool, optional
        Use openbabel and structure information to estimate a charge/spin state for the molecule
    charge : int, optional 
        charge on the molecule, default None
    uhf : int, optional
        unpaired electrons in the molecule, default None
    xtb_uhf : int, optional
        unpaired electrons in molecule for use by XTB, default None
    xtb_charge : int, optional
        charge used for XTB relaxation, default None
    """
    mol = Molecule()
    if isinstance(structure,(str,ase.atoms.Atoms)):
        # Mol2string
        if 'TRIPOS' in structure:
            # mol = structure
            mol.read_mol2(structure,readstring=True)
        # Xyz filename
        elif structure[-4:] == '.xyz':
            mol.read_xyz(structure,readstring=False)
        # rxyz filename
        elif structure[-5:] == '.rxyz':
            mol.read_rxyz(structure)
        # mol2 filename
        elif structure[-5:] == '.mol2':
            mol.read_mol2(structure,readstring=False)
        elif isinstance(structure,str) and (len(structure.split('\n')) > 3) and (structure.split('\n')[0].replace(' ','').isnumeric()) \
            and ('FORCES' in structure) and ('ENERGY' in structure): # RXYZ string.
            mol.read_rxyz(structure,readstring=True)
        # checking for number at start of string -> indicates xyz string
        elif isinstance(structure,str) and (len(structure.split('\n')) > 3) and (structure.split('\n')[0].replace(' ','').isnumeric()):
            mol.read_xyz(structure,readstring=True)
        # checking for similar file without header
        elif isinstance(structure,str) and (len(structure.split('\n')[0].split()) == 4) and structure.split('\n')[0].split()[0]:
            mol.read_xyz(structure,readstring=True)
        elif isinstance(structure,str): # Smiles?
            try:
                tmol = io_obabel.get_obmol_smiles(structure)
                mol2 = io_obabel.convert_obmol_mol2(tmol)
                mol.read_mol2(mol2,readstring=True)
                mol.charge = tmol.GetTotalCharge()
                charges = np.array(np.zeros(len(mol.ase_atoms)))
                charges[0] = mol.charge
                mol.ase_atoms.set_initial_charges(charges)
            except:
                raise ValueError('Not Recognized Structure Type (str)!')
        elif isinstance(structure, ase.atoms.Atoms):
            mol.load_ase(structure,atom_types=structure.get_chemical_symbols())
        else:
            raise ValueError('Not Recognized Structure Type!')
        if detect_charge_spin:
            mol.detect_charge_spin()
            return mol
        else:
            if charge is not None:
                mol.charge = charge
                charges = np.zeros(len(mol.ase_atoms))
                charges[0] = mol.charge
                mol.ase_atoms.set_initial_charges(charges)
            if uhf is not None:
                mol.uhf = uhf
                uhf_vect = np.zeros(len(mol.ase_atoms))
                mol.ase_atoms.set_initial_magnetic_moments(uhf_vect)
            if xtb_uhf is not None:
                mol.xtb_uhf = xtb_uhf
            if xtb_charge is not None:
                mol.xtb_charge = xtb_charge
            return mol
    elif isinstance(structure,architector.io_molecule.Molecule):
        # struct = mol
        # for key,val in structure.__dict__.items():
        #         setattr(struct,key,val)
        structure.ase_atoms.calc = None
        struct = copy.deepcopy(structure)
        if detect_charge_spin:
            struct.detect_charge_spin()
            return struct
        else:
            if charge is not None:
                struct.charge = charge
                charges = np.zeros(len(struct.ase_atoms))
                charges[0] = struct.charge
                struct.ase_atoms.set_initial_charges(charges)
            if uhf is not None:
                struct.uhf = uhf
                uhf_vect = np.zeros(len(struct.ase_atoms))
                struct.ase_atoms.set_initial_magnetic_moments(uhf_vect)
            if xtb_uhf is not None:
                mol.xtb_uhf = xtb_uhf
            return struct
    else:
        raise ValueError('String needed for classifying structure type.')

def convert_ase_xyz(ase_atoms):
    """convert_ase_xyz

    Parameters
    ----------
    ase_atoms : ase.Atoms
        ase atoms to write to xyz string

    Returns
    -------
    outstring : str
        xyz file string
    """
    outstring = '{}\n\n'.format(len(ase_atoms))
    for atom in ase_atoms:
        outstring += '{} {} {} {}\n'.format(atom.symbol,
                                            atom.position[0],
                                            atom.position[1],
                                            atom.position[2])
    outstring = outstring.strip('\n')
    return outstring


def convert_xyz_ase(structure_str):
    """convert_xyz_ase

    Parameters
    ----------
    structure_str : str
        xyz file string

    Returns
    -------
    ase_atoms : ase.Atoms
        ase atoms to write to xyz string
    """
    return ase.io.read(StringIO(structure_str), format="xyz")


class Molecule:

    def __init__(self, in_ase=False, BO_dict={}, atom_types=[], cell=[], charge=None,
                uhf=None, xtb_uhf=None, xtb_charge=None):
        self.dists_sane = True
        self.sanity_check_dict = {}
        if isinstance(in_ase,ase.atoms.Atoms):
            self.ase_atoms = in_ase.copy()
            self.BO_dict = BO_dict
            self.charge = charge
            self.uhf = uhf
            self.xtb_uhf = xtb_uhf
            self.xtb_charge = xtb_charge
            self.atom_types = atom_types
            self.cell = cell
            if len(BO_dict) > 0:
                self.graph = np.zeros((len(self.ase_atoms),len(self.ase_atoms)))
                for key,_ in self.BO_dict.items():
                    i = int(key[0]) - 1 # BO Dict is 1-index (thanks to  OBmol/mol2 format)
                    j = int(key[1]) - 1
                    self.graph[i,j] = 1 
                    self.graph[j,i] = 1 
            else:
                self.graph = []
        elif isinstance(in_ase,str) and (in_ase[-4:] == '.xyz'):
            self.read_xyz(in_ase)
            self.BO_dict = BO_dict
            self.atom_types = atom_types
            self.charge = charge
            self.uhf = uhf
            self.xtb_uhf = xtb_uhf
            self.xtb_charge = xtb_charge
            self.cell = cell
            if len(BO_dict) > 0:
                self.graph = np.zeros((len(self.ase_atoms),len(self.ase_atoms)))
                for key,_ in self.BO_dict.items():
                    i = int(key[0]) - 1 # BO Dict is 1-index (thanks to  OBmol/mol2 format)
                    j = int(key[1]) - 1
                    self.graph[i,j] = 1 
                    self.graph[j,i] = 1 
            else:
                self.graph = []
        elif isinstance(in_ase,str) and (in_ase[-5:] == '.mol2'):
            self.read_mol2(in_ase)
        elif isinstance(in_ase,bool):
            self.ase_atoms = None
        else:
            raise ValueError('Need ase.atoms.Atoms/xyz/mol2 as input for molecule class!')

    def load_ase(self,in_ase, BO_dict=dict(), atom_types=[], cell=[],
                charge=None, uhf=None, xtb_uhf=None, xtb_charge=None):
        """load_ase read in ase atoms object.

        Parameters
        ----------
        in_ase : ase.atoms.Atoms
            structure
        BO_dict : dict, optional
            bond order dictionary, by default dict()
        atom_types : list, optional
            list of atom types (ideally sybyl style), by default []
        cell : list, optional
            unit cell, by default []
        charge : int/float, optional
            charge on the system, by default None
        uhf : int/float, optional
            number unpaired electrons on the system, by default None
        xtb_uhf : int/float, optional
            number unpaired electrons on the system desired for XTB, by default None

        Raises
        ------
        ValueError
            If ase atoms not passed.
        """
        if isinstance(in_ase,ase.atoms.Atoms):
            self.ase_atoms = in_ase.copy()
        else:
            raise ValueError('Need ase.atoms.Atoms as input for molecule class!')
        self.BO_dict = BO_dict.copy()
        self.atom_types = atom_types
        self.cell = cell
        if charge is not None:
            self.charge = charge
        else:
            self.charge = np.sum(self.ase_atoms.get_initial_charges())
        if uhf is not None:
            self.uhf = uhf
        else:
            self.uhf = np.sum(self.ase_atoms.get_initial_magnetic_moments())
        if xtb_uhf is not None:
            self.xtb_uhf = xtb_uhf
        else:
            self.xtb_uhf = int(np.sum(self.ase_atoms.get_initial_magnetic_moments()))
        if xtb_charge is not None:
            self.xtb_charge = xtb_charge
        else: 
            self.xtb_charge = np.sum(self.ase_atoms.get_initial_charges())
        if len(self.BO_dict) > 0:
            self.graph = np.zeros((len(self.ase_atoms),len(self.ase_atoms)))
            for key,_ in self.BO_dict.items():
                i = int(key[0]) - 1 # BO Dict is 1-index (thanks to  OBmol/mol2 format)
                j = int(key[1]) - 1
                self.graph[i,j] = 1 
                self.graph[j,i] = 1 
        else:
            self.graph = []
            
    def write_xyz(self, filename, writestring=False):
        """convert_ase_xyz

        Parameters
        ----------
        ase_atoms : ase.Atoms
            ase atoms to write to xyz string

        Returns
        -------
        outstring : str
            xyz file string
        """
        ase_atoms = self.ase_atoms
        outstring = '{}\n\n'.format(len(ase_atoms))
        for atom in ase_atoms:
            outstring += '{} {} {} {}\n'.format(atom.symbol,
                                                atom.position[0],
                                                atom.position[1],
                                                atom.position[2])
        outstring = outstring.strip('\n')
        if writestring:
            return outstring
        else:
            if filename[-4:] == '.xyz':
                filename = filename
            else:
                filename = filename.replace('.','') + '.xyz'
            with open(filename,'w') as file1:
                file1.write(outstring)

    def read_xyz(self,filename,readstring=False):
        """read_xyz read in an xyz file

        Parameters
        ----------
        filename : str
            name of file to read from
        readstring : bool, optional
            whether to read from a string, by default False
        """
        if readstring:
            filename = StringIO(filename)
            self.ase_atoms = read(filename,format='xyz',parallel=False)
        else:
            self.ase_atoms = read(filename,parallel=False)
        self.graph = []
        self.BO_dict = {}
        self.charge = None
        self.uhf = None
        self.xtb_charge = None
        self.xtb_uhf = None
        self.atom_types = [x.symbol for x in self.ase_atoms]
        self.cell = []

    def read_rxyz(self,filename,readstring=False):
        """read_rxyz read an RXYZ file
        - Specific to LANL

        Parameters
        ----------
        filename : str
            name of file to read from
        readstring : bool, optional
            whether to read from a string, by default False
        """
        if not readstring:
            with open(filename,'r') as file1:
                lines = file1.readlines()
        else:
            lines = filename.split('\n')
        atoms = []
        for line in lines:
            sline = line.split()
            if 'FORCES' in line:
                break
            elif len(sline) == 4:
                symbol = sline[0]
                coords = (float(sline[1]),float(sline[2]),float(sline[3]))
                if len(atoms) == 0:
                    atoms = Atoms([Atom(symbol,coords)])
                else:
                    atoms.append(Atom(symbol,coords))
        self.ase_atoms = atoms
        self.graph = []
        self.charge = None
        self.uhf = None
        self.xtb_charge = None
        self.xtb_uhf = None
        self.BO_dict = {}
        self.atom_types = [x.symbol for x in self.ase_atoms]
        self.cell = []

    def write_mol2(self, filename, writestring=False):
        """write_mol2 routine for writing a mol2 file.
        Works with integrating with CSD software and openbabel.

        Parameters
        ----------
        filename : str
            name of file to read from
        writestring : bool, optional
            whether to write to string, by default False

        Returns
        -------
        ss : str
            mol2 file in string format
        """
        if len(self.graph) < 1:
            self.create_mol_graph()
        natoms = len(self.ase_atoms)
        csg = csgraph_from_dense(self.graph)
        disjoint_components = connected_components(csg)
        if disjoint_components[0] > 1:
            atom_group_names = ['RES'+str(x+1) for x in disjoint_components[1]]
            atom_groups = [str(x+1) for x in disjoint_components[1]]
        else:
            atom_group_names = ['RES1']*natoms
            atom_groups = [str(1)]*natoms
        charges = np.zeros(natoms)
        charge_string = 'NoCharges'
        if (self.charge is not None) and (self.uhf is not None) and (self.xtb_charge is not None) and (self.xtb_uhf is not None):
            ss = '@<TRIPOS>MOLECULE\n{} Charge: {} Unpaired_Electrons: {} XTB_Unpaired_Electrons: {} XTB_Charge: {}\n'.format(filename,
                    int(self.charge),int(self.uhf),int(self.xtb_uhf),int(self.xtb_charge))
        else:
            ss = '@<TRIPOS>MOLECULE\n{}\n'.format(filename)            
        ss += ' {0:5d} {1:5d} {2:5d} {3:5d} {4:5d}\n'.format(natoms,
                                    int(csg.nnz/2), disjoint_components[0],
                                    0, 0)
        ss += 'SMALL\n'
        ss += charge_string + '\n' + '****\n' + 'Generated from Architector\n\n'
        ss += '@<TRIPOS>ATOM\n'
        atom_default_dict = {'C': '3', 'N': '3', 'O': '2', 'S': '3', 'P': '3'}
        atom_types = self.ase_atoms.get_chemical_symbols()
        atom_type_numbers = np.ones(len(atom_types))
        atom_types_mol2 = []
        for i, atom in enumerate(self.ase_atoms):
            if atom.symbol != self.atom_types[i]:
                atom_types_mol2 = self.atom_types[i]
            elif atom.symbol in list(atom_default_dict.keys()):
                atom_types_mol2 = atom.symbol + '.' + atom_default_dict[atom.symbol]
            else:
                atom_types_mol2 = atom.symbol
            type_ind = atom_types.index(atom.symbol)
            atom_coords = atom.position
            ss += '{0:6d} {1:6s} {2:9.4f} {3:9.4f} {4:9.4f}   {5:6s}{6:5d} {7:5s}{8:8.4f}\n'.format(
                i+1, atom.symbol+str(int(atom_type_numbers[type_ind])), 
                atom_coords[0], atom_coords[1], atom_coords[2], 
                atom_types_mol2, int(atom_groups[i]),
                atom_group_names[i], charges[i]
            )
            atom_type_numbers[type_ind] += 1
        ss += '@<TRIPOS>BOND\n'
        bonds = csg.nonzero()
        bond_count = 1
        if self.BO_dict:
            bondorders = True
        else:
            bondorders = False
        for i, b1 in enumerate(bonds[0]):
            b2 = bonds[1][i]
            if b2 > b1 and not bondorders:
                ss += '{0:6d}{1:6d}{2:6d}{3:>5s}\n'.format(
                    bond_count, b1+1, b2+1, str(1)
                )
                bond_count += 1
            elif b2 > b1 and bondorders:
                ss += '{0:6d}{1:6d}{2:6d}{3:>5s}\n'.format(
                    bond_count, b1+1, b2+1, str(self.BO_dict[(int(b1)+1, int(b2)+1)])
                )
                bond_count += 1
        ss += '@<TRIPOS>SUBSTRUCTURE\n'
        unique_group_names = np.unique(atom_group_names)
        for i, name in enumerate(unique_group_names):
            ss += '{0:6d} {1:6s}{2:7d} GROUP             0 ****  ****    0  \n'.format(
                i+1, name, atom_group_names.count(name) 
            )

        if hasattr(self,'cell'):
            # If there's a cell, write it - else 'ehh'
            if (len(self.cell) == 6): #Only abc,alphabetagamma defined (add 1 1 for space group)
                ss += '@<TRIPOS>CRYSIN\n'
                ss += '{0:10.4f}{1:10.4f}{2:10.4f}{3:10.4f}{4:10.4f}{5:10.4f}{6:6d}{7:6d}\n'.format(
                    self.cell[0], self.cell[1], self.cell[2],
                    self.cell[3], self.cell[4], self.cell[5],
                    '1', '1'
                    )
            elif (len(self.cell) == 8): #abc,alphabetagamma and space group defined
                ss += '@<TRIPOS>CRYSIN\n'
                ss += '{0:10.4f}{1:10.4f}{2:10.4f}{3:10.4f}{4:10.4f}{5:10.4f}{6:6d}{7:6d}\n'.format(
                    self.cell[0], self.cell[1], self.cell[2],
                    self.cell[3], self.cell[4], self.cell[5],
                    self.cell[6], self.cell[7]
                    )

        if writestring:
            return ss
        else:
            if '.mol2' not in filename:
                if '.' not in filename:
                    filename += '.mol2'
                else:
                    filename = filename.split('.')[0]+'.mol2'
            with open(filename, 'w') as file1:
                file1.write(ss)

    def read_mol2(self, filename, readstring=False):
        """ Read mol2 into a Mol class instance. Stores the bond orders and atom types (SYBYL).
        Will read in charge and spin states assigned to complexes generated from architector to the molecule!

        Parameters
        -------
        filename : string
            String of path to XYZ file. Path may be local or global. May be read in as a string.
        readstring : bool, optional
            Flag for deciding whether a string of mol2 file is being passed as the filename, default False
        """
        # ptable=PTable()
        graph = False
        bo_dict = False
        self.atom_types = []
        if readstring:
            s = filename.splitlines()
        else:
            with open(filename, 'r') as f:
                s = f.read().splitlines()
        read_atoms = False
        read_bonds = False
        read_cell = False
        charge = None
        spin = None
        xtb_spin = None
        xtb_charge = None
        for line in s:
            # Get Atoms First
            if ('Charge:' in line) and ('Unpaired_Electrons:' in line):
                charge = float(line.split()[2])
                spin = int(line.split()[4])
                xtb_spin = int(line.split()[6])
                xtb_charge = int(line.split()[8])
            if ('<TRIPOS>BOND' in line) or ('<TRIPOS>UNITY_ATOM_ATTR' in line):
                read_atoms = False
            if ('<TRIPOS>SUBSTRUCTURE' in line) or ('<TRIPOS>UNITY_ATOM_ATTR' in line):
                read_bonds = False
                read_atoms = False
            if read_atoms:
                s_line = line.split()
                # Check redundancy in Chemical Symbols
                atom_symbol1 = re.sub('[0-9]+[A-Z]+', '', line.split()[1])
                atom_symbol1 = re.sub('[0-9]+', '', atom_symbol1)
                atom_symbol2 = line.split()[5]
                if len(atom_symbol2.split('.')) > 1:
                    atype = atom_symbol2
                else:
                    atype = False
                atom_symbol2 = atom_symbol2.split('.')[0]
                if atom_symbol1 in io_ptable.elements:
                    atom = ase.atom.Atom(atom_symbol1, [float(s_line[2]), float(
                        s_line[3]), float(s_line[4])])
                    if isinstance(atype,str):
                        self.atom_types.append(atype)
                    else:
                        self.atom_types.append(atom_symbol1)
                elif atom_symbol2 in io_ptable.elements:
                    atom = ase.atom.Atom(atom_symbol2, [float(s_line[2]), float(
                        s_line[3]), float(s_line[4])])
                    if isinstance(atype,str):
                        self.atom_types.append(atype)
                    else:
                        self.atom_types.append(atom_symbol2)
                else:
                    raise ValueError('Cannot find atom symbol in ptable')
                if isinstance(self.ase_atoms,ase.atoms.Atoms):
                    self.ase_atoms.append(atom)
                else:
                    self.ase_atoms = ase.atoms.Atoms([atom])
            if read_bonds:  # Read in bonds to molecular graph
                s_line = line.split()
                graph[int(s_line[1]) - 1, int(s_line[2]) - 1] = 1
                graph[int(s_line[2]) - 1, int(s_line[1]) - 1] = 1
                bo_dict[tuple(
                    sorted([int(s_line[1]), int(s_line[2])]))] = s_line[3]
            if read_cell:
                s_line = line.split()
                self.cell = [float(s_line[0]), float(s_line[1]), float(s_line[2]),
                        float(s_line[3]), float(s_line[4]), float(s_line[5]),
                        int(s_line[6]),int(s_line[7])]
            if '<TRIPOS>ATOM' in line:
                read_atoms = True
            if '<TRIPOS>BOND' in line:
                read_bonds = True
                # initialize molecular graph
                graph = np.zeros((len(self.ase_atoms), len(self.ase_atoms)))
                bo_dict = dict()
            if '<TRIPOS>CRYSIN' in line:
                read_cell = True
        if isinstance(graph, np.ndarray):  # Enforce mol2 molecular graph if it exists
            self.graph = graph
            self.BO_dict = bo_dict
        else:
            self.graph = []
            self.BO_dict = {}
        self.charge = charge
        self.uhf = spin
        self.xtb_uhf = xtb_spin
        self.xtb_charge = xtb_charge

    def detect_charge_spin(self):
        """detect_charge_spin 
        Detect the spin/charge of the complex using openbabel intepretations of bond orders
        and partial charges. Definitely do not guarantee this is correct for most cases.
        """
        metals = np.array(self.find_metals())
        # Metal charges
        if len(metals) > 0:
            charge = np.sum([io_ptable.metal_charge_dict[x] for x in np.array(self.ase_atoms.get_chemical_symbols())[metals]])
            mol2str = self.write_mol2('thing.mol2',writestring=True)
            _, _, infodict = io_obabel.obmol_lig_split(mol2str,return_info=True,calc_coord_atoms=False)
            charge += np.sum(infodict['lig_charges'])
            met_syms = np.array(self.ase_atoms.get_chemical_symbols())[metals]
            uhf = np.sum([io_ptable.metal_spin_dict[x] for x in met_syms])
        else:
            raise ValueError('Detect_charge_spin not implemented for non-metal containing systems.')
        even_odd_electrons = (np.sum([atom.number for atom in self.ase_atoms])-charge) % 2
        if (uhf is not None):
            uhf = uhf
            if (even_odd_electrons == 1) and (uhf == 0):
                uhf = 1
            elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
                uhf += 1
            elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
                uhf -= 1
            if (even_odd_electrons == 0) and (uhf % 2 == 1):
                uhf = uhf - 1 
            elif (even_odd_electrons == 1) and (uhf % 2 == 0):
                uhf = uhf + 1
        xtb_uhf = copy.deepcopy(uhf)
        if np.any([True for x in met_syms if x in io_ptable.heavy_metals]):
            xtb_uhf = 0
            if (even_odd_electrons == 1):
                xtb_uhf = 1
        self.charge = int(charge)
        self.uhf = int(uhf)
        self.xtb_uhf = int(xtb_uhf)
        self.xtb_charge = int(charge)

    def find_metal(self):
        """find_metal 
        pull out the metal index in a complex

        Returns
        -------
        metalind : int
            index of the metal or first metal if present.
        """
        syms = self.ase_atoms.get_chemical_symbols()
        metalinds = [i for i,x in enumerate(syms) if x in io_ptable.all_metals]
        if len(metalinds) == 1:
            metalind = metalinds[0]
        elif len(metalinds) > 1:
            print('Assigning first metal as metalind.')
            metalind = metalinds[0]
        else:
            print('No metals in this molecule.')
            metalind = None
        return metalind

    def find_metals(self):
        """find_metals find all metals in a molecule

        Returns
        -------
        metals : list
            indices of all the metals in the molecule
        """
        metals = [i for i,x in enumerate(self.ase_atoms) if (x.symbol in io_ptable.all_metals)]
        return metals

    def remove_atom(self,ind):
        """remove_atom delete an atom from the molecule with ind

        Parameters
        ----------
        ind : int
            index to remove
        """
        del self.ase_atoms[ind]
        self.graph = np.delete(np.delete(self.graph,ind,0),ind,1)
        del self.atom_types[ind]
        BO_dict_tmp = {x:y for x,y in self.BO_dict.items() if (ind+1 not in x)}
        bo_dict_out = dict()
        for x,y in BO_dict_tmp.items():
            newx = []
            if x[0] > ind+1:
                newx.append(x[0] - 1)
            else:
                newx.append(x[0])
            if x[1] > ind+1:
                newx.append(x[1] - 1)
            else:
                newx.append(x[1])
            newx = tuple(newx)
            bo_dict_out[newx] = y
        self.BO_dict = bo_dict_out

    def remove_metals(self):
        """remove_metals 
        remove all metals from the molecule
        """
        metals = self.find_metals()
        for i in sorted(metals)[::-1]:
            self.remove_atom(i)
                
    def create_mol_graph(self, cutoffs=True, skin=0.2, allow_mm_bonds=False):
        """create_mol_graph 
        Create molecular graph based on cutoffs, default skin=0.2

        Parameters
        ----------
        cutoffs : bool, optional
            pass in a list of cutoffs for the structure?, by default True
        skin : float, optional
            tolerance for considering neighbors, by default 0.2
        allow_mm_bonds : bool, optional
            allow metal-metal bonds?, by default False
        """
        if cutoffs:
            cutoffsvect=np.array([io_ptable.rcov1[atom.number] for atom in self.ase_atoms])
        metals = self.find_metals()
        # Broadcast cutoffs sum
        cutoff_dist_mat = (cutoffsvect + skin) + (cutoffsvect+skin)[:,None]
        graph = np.zeros((len(self.ase_atoms), len(self.ase_atoms)))
        coords = self.ase_atoms.get_positions()
        # Calculate interatomic distances.
        act_dist_mat = np.linalg.norm(coords[:,None,:]-coords[None,:,:],axis=-1)
        delta_dist_mat = act_dist_mat - cutoff_dist_mat
        graph[np.where(delta_dist_mat < 0)] = 1
        graph = graph-np.eye(len(self.ase_atoms))
        if (len(metals) > 0):
            if (len(metals) > 1) and (not allow_mm_bonds):
                for i,j in itertools.combinations(metals,2):
                    graph[i,j] = 0  
                    graph[j,i] = 0

        # # Remove self interaction/bonding
        self.graph = graph
        self.create_BO_dict(cutoffs=cutoffs,skin=skin)

    def create_BO_dict(self, cutoffs=False, skin=0.2):
        """create_BO_dict
        Create BO dict - default everything to bond order 1.

        Parameters
        ----------
        cutoffs : bool, optional
            pass in a list of cutoffs for the structure?, by default True
        skin : float, optional
            tolerance for considering neighbors, by default 0.2
        """
        if len(self.graph) < 1:
            self.create_mol_graph(cutoffs=cutoffs, skin=skin)
        csg = csgraph_from_dense(self.graph)
        bonds = csg.nonzero()
        bond_dict = dict()
        for i, b1 in enumerate(bonds[0]):
            b2 = bonds[1][i]
            if b2 > b1:
                bond_dict[(b1+1,b2+1)] = '1'
        self.BO_dict = bond_dict

    def create_graph_from_bo_dict(self):
        """create_graph_from_bo_dict 
        Routine to create a molecular graph directly from a dictionary of bond orders
        """
        if len(self.BO_dict) > 0:
            self.graph = np.zeros((len(self.ase_atoms),len(self.ase_atoms)))
            for key,_ in self.BO_dict.items():
                i = int(key[0]) - 1 # BO Dict is 1-index (thanks to  OBmol/mol2 format)
                j = int(key[1]) - 1
                self.graph[i,j] = 1 
                self.graph[j,i] = 1 
        else:
            self.graph = []

    def append_ligand(self, ligand=None):
        """append_ligand Add ligand to the structure
        Account for charges on the ligand.

        Parameters
        ----------
        ligand : dict, optional
            ligand to add to complex, by default None
        """
        lig_bo_dict = ligand['bo_dict']
        lig_ase_atoms = ligand['ase_atoms']
        lig_atom_types = ligand['atom_types']
        natoms = len(self.ase_atoms)
        newligbodict = dict()
        for key,val in lig_bo_dict.items():
            newkey = [0,0]
            if int(key[0]) > 1:
                newkey[0] = natoms + int(key[0]) - 1
            else:
                newkey[0] = 1
            if int(key[1]) > 1:
                newkey[1] = int(key[1]) + natoms - 1
            else:
                newkey[1] = 1
            newkey = tuple(newkey)
            newligbodict.update({newkey:val})
        self.BO_dict.update(newligbodict)
        lcs = lig_ase_atoms.get_initial_charges().sum()
        self.charge += lcs
        self.xtb_charge += lcs
        self.ase_atoms += lig_ase_atoms
        self.atom_types += lig_atom_types
        self.create_graph_from_bo_dict()

    def graph_sanity_checks(self, factor=1.45, params={}, assembly=False):
        """graph_sanity_checks
        Check if any part of the molecule is blown up relative to the imposed molecular graph

        Parameters
        ----------
        factor : float, optional
            tolerance for long bonds - factor*(sum of covalent radii), by default 1.4
        params : dict, optional
            parameters from inputDict, default {}
        assembly : bool, optional
            if this is an assembly check or final check, default False -> final cutoffs.

        Returns
        -------
        sane : bool
            If the graph distances are longer than the factor -> indicating relatively garbage
            geometry.
        """
        run_check = True
        if (len(params) > 0) and (not assembly):
            run_check = params['full_sanity_checks']
            factor = params['full_graph_sanity_cutoff']
        elif (len(params) > 0):
            run_check = params['assemble_sanity_checks']
            factor = params['assemble_graph_sanity_cutoff']
        else:
            params = {'covrad_metal':None}
        sane = self.dists_sane
        graph_dists_dict = {}
        if run_check:
            if len(self.ase_atoms) > 1: # Don't test distances for single atom
                posits = self.ase_atoms.get_positions()
                m_inds = [i for i,x in enumerate(self.ase_atoms.get_chemical_symbols()) if x in io_ptable.all_metals]
                m_ind=None
                if len(m_inds) > 1:
                    raise ValueError('Sanity check only implemented for mononuclear so far.')
                elif len(m_inds) == 1:
                    m_ind = m_inds[0]
                mrad = None
                if isinstance(params['covrad_metal'],float):
                    mrad = params['covrad_metal']
                if np.any(np.isnan(posits)): # Any nan in positions
                    sane = False
                    graph_dists_dict.update({'containsNAN_posit':True})
                else:
                    all_dists = self.ase_atoms.get_all_distances()
                    cov_radii = np.array([io_ptable.rcov1[x] for x in self.ase_atoms.get_atomic_numbers()]) 
                    if (not (m_ind is None)) and (not (mrad is None)):
                        if mrad >= cov_radii[m_ind]:
                            cov_radii[m_ind] = mrad
                    for key, _ in self.BO_dict.items():
                        i = key[0] - 1
                        j = key[1] - 1
                        if all_dists[i,j] > factor*(cov_radii[i] + cov_radii[j]):
                            sane = False
                            graph_dists_dict.update({'Cutoff':factor})
                            graph_dists_dict.update({(i,j):all_dists[i,j]/(cov_radii[i] + cov_radii[j])})
                            if params.get('debug',False):
                                print('Graph distance long: ', all_dists[i,j]/(cov_radii[i] + cov_radii[j]))
                            break
        self.dists_sane = sane
        self.sanity_check_dict.update({'Graph_Dist_Checks':graph_dists_dict})

    def dist_sanity_checks(self,
                 smallest_dist_cutoff=0.55,
                 min_dist_cutoff=3,
                 params={},
                 assembly=False,
                 covrad_metal=None,
                 debug=False):
        """dist_sanity_checks
        Perform basic distance-based sanity checks on structure

        Parameters
        -------
        atoms : ase.Atoms
            atoms to check for sanity.
        params : dict, optional
            parameters for dictionary, default {}.
        smallest_dist_cutoff : float
            distance cutoff-make sure sum of cov radii larger than dist*smallest_dist_cutoff
        min_dist_cutoff : int/float
            make sure all atoms are at least min_dist_cutoff from ANY other atom
        assembly : bool
            whether this is an assembly step or final relaxation
        covrad_metal : float
            the covalent radii of the metal if requested
        debug : bool
            print if debug requested.
        """
        run_check = True
        if (len(params) > 0)  and (not assembly):
            run_check = params['full_sanity_checks']
            smallest_dist_cutoff = params['full_smallest_dist_cutoff']
            min_dist_cutoff = params['full_min_dist_cutoff']
        elif (len(params) > 0) :
            run_check = params['assemble_sanity_checks']
            smallest_dist_cutoff = params['assemble_smallest_dist_cutoff']
            min_dist_cutoff = params['assemble_min_dist_cutoff']
        sane = self.dists_sane
        min_dist_dict = {}
        smallest_dist_dict = {}
        if run_check:
            atoms = self.ase_atoms.copy()
            if len(atoms) > 1: # Don't test distances for single atom
                posits = atoms.get_positions()
                m_inds = [i for i,x in enumerate(atoms.get_chemical_symbols()) if x in io_ptable.all_metals]
                m_ind=None
                if len(m_inds) > 1:
                    raise ValueError('Sanity check only implemented for mononuclear so far.')
                elif len(m_inds) == 1:
                    m_ind = m_inds[0]
                mrad = None
                if isinstance(covrad_metal,float):
                    mrad = covrad_metal
                elif isinstance(params,dict):
                    if isinstance(params.get('covrad_metal',False),float):
                        mrad = params['covrad_metal']
                if np.any(np.isnan(posits)): # Any nan in positions
                    if params.get('debug',False):
                        print('Nan in positions.')
                    sane = False
                else:
                    all_dists = atoms.get_all_distances()
                    cov_radii = np.array([io_ptable.rcov1[x] for x in atoms.get_atomic_numbers()])
                    if (not (m_ind is None)) and (not (mrad is None)):
                        cov_radii[m_ind] = mrad
                    for i in range(0,len(atoms)):
                        j_list = list(range(0,len(atoms)))
                        j_list.remove(i)
                        i_dists = []
                        for j in j_list:
                            i_dists.append(all_dists[i,j])
                            if all_dists[i,j] < smallest_dist_cutoff*(cov_radii[i] + cov_radii[j]): # Check for extra crowded metals
                                sane = False
                                smallest_dist_dict.update({'Cutoff':smallest_dist_cutoff})
                                smallest_dist_dict.update({(i,j):all_dists[i,j]/(cov_radii[i] + cov_radii[j])})
                                if params.get('debug',False):
                                    print('Dist short: ', all_dists[i,j])
                        if min(i_dists) > min_dist_cutoff: # Catch cases where atom shot off metal center or blown up structure
                            sane = False
                            min_dist_dict.update({'Cutoff':min_dist_cutoff})
                            min_dist_dict.update({i:min(i_dists)})
                            if params.get('debug',False):
                                print('Mindist long: ', min(i_dists))
        self.dists_sane = sane
        self.sanity_check_dict.update({'Smallest_Dist_Checks':smallest_dist_dict, 'Min_Dist_Checks':min_dist_dict})

    def get_can_label(self):
        """ Get molecular graph determinant - serves as unique identifier. 

        Returns
        -------
        safedet : str
            String containing the molecular graph determinant.
        """
        if not len(self.graph):
            self.create_mol_graph()
        syms = self.ase_atoms.get_chemical_symbols()
        weights = [io_ptable.masses[io_ptable.elements.index(x)] for x in syms]
        inds = np.nonzero(self.graph)
        tmpgraph = self.graph.copy()
        for j in range(len(syms)): # Set diagonal to weights
            tmpgraph[j, j] = weights[j]
        for i, x in enumerate(inds[0]):
            y = inds[1][i]
            # Add factor of 100
            tmpgraph[x, y] = weights[x]*weights[y]*tmpgraph[x, y] / 100.0
        with np.errstate(over='raise'):
            try:
                det = np.linalg.det(tmpgraph)
            except:
                det = np.linalg.det(tmpgraph/100.0)
        if 'e+' in str(det):
            safedet = str(det).split(
                'e+')[0][0:10]+'e+'+str(det).split('e+')[1]
        else:
            safedet = str(det)[0:10]
        return safedet

    def calc_suggested_spin(self,params={}):
        """calc_suggested_spin calculate the suggested spin using electron counting and given information.

        Parameters
        ----------
        params : dict, optional
            inputDict parameters , by default dict()
        """
        # Charge -> charges already assigned to components during assembly
        metals = self.find_metals()
        charge_vect = np.zeros(len(self.ase_atoms))
        # Prioritize input full charge > ase atoms charge > self.charge > metal_ox > None
        if (params.get('full_charge',None) is not None):
            charge_vect[0] = params['full_charge']
        elif (self.charge is not None) and (self.xtb_charge != self.charge):
            charge_vect[0] = self.charge
        elif np.any(self.ase_atoms.get_initial_charges() != 0): # Prioritize ASE atoms charge
            charge_vect = self.ase_atoms.get_initial_charges()
        elif self.charge is not None: 
            charge_vect[0] = self.charge
        else:
            if (params.get('metal_ox', None) is not None):
                charge_vect[0] = params['metal_ox']
            else:
                if len(metals) > 0:
                    syms = self.ase_atoms.get_chemical_symbols()
                    charge_vect[0] = np.sum([io_ptable.metal_charge_dict[syms[x]] for x in metals])
                else:
                    mol2str = self.write_mol2('cool.mol2',writestring=True)
                    tmol = io_obabel.convert_mol2_obmol(mol2str,readstring=True)
                    charge_vect[0] = tmol.GetTotalCharge()
                    
        mol_charge = np.sum(charge_vect)
        xtb_charge = copy.deepcopy(mol_charge)
        if np.any([True for x in self.ase_atoms.get_chemical_symbols() if x in io_ptable.heavy_metals]) and \
            ((params.get('metal_ox', None) is not None)):
            xtb_charge = mol_charge + (3-params.get('metal_ox', None))

        # Handle spin / magnetism
        even_odd_electrons = (np.sum([atom.number for atom in self.ase_atoms])-mol_charge) % 2
        # Prioritize full spin > metal spin > self.uhf > default metal spin
        if(params.get('full_spin', None) is None):
            if params.get('metal_spin',None) is None:
                if self.uhf is None:
                    syms = self.ase_atoms.get_chemical_symbols()
                    # If no metals -> uhf will start at 0
                    uhf = np.sum([io_ptable.metal_spin_dict[syms[x]] for x in metals])
                else:
                    uhf = self.uhf
            else:
                uhf = params['metal_spin'] # Metal spin set by io_process_input to defaults.
        else:
            uhf = params['full_spin']

        # Only correct uhf if anything is wrong with uhf -> even vs odd etc.
        if (even_odd_electrons == 1) and (uhf == 0):
            uhf = 1
        elif (even_odd_electrons == 1) and (uhf < 7) and (uhf % 2 == 0):
            uhf += 1
        elif (even_odd_electrons == 1) and (uhf >= 7) and (uhf % 2 == 0):
            uhf -= 1
        if (even_odd_electrons == 0) and (uhf % 2 == 1):
            uhf = uhf - 1 
        elif (even_odd_electrons == 1) and (uhf % 2 == 0):
            uhf = uhf + 1

        xtb_uhf = 0
        if not np.any([True for x in self.ase_atoms.get_chemical_symbols() if x in io_ptable.heavy_metals]):
            xtb_uhf = uhf
        else: # F in core assumes for a 3+ lanthanide there are 11 valence electrons (8 once the 3+ is taken into account)
            even_odd_electrons = (np.sum([atom.number for atom in self.ase_atoms]))
            even_odd_electrons = even_odd_electrons - \
                io_ptable.elements.index(self.ase_atoms.get_chemical_symbols()[metals[0]]) + 11 - xtb_charge
            even_odd_electrons = even_odd_electrons % 2
            if (even_odd_electrons == 0):
                xtb_uhf = 0
            else:
                xtb_uhf = 1
        
        # Assign to complex
        self.charge = mol_charge
        self.uhf = uhf
        self.xtb_uhf = xtb_uhf
        self.xtb_charge = xtb_charge



import tempfile
import shutil
import contextlib
import os

@contextlib.contextmanager
def make_temp_directory(prefix=None):
    """make_temp_directory function to make a temporary directory and change there.
    Very useful for running on supercomputers with scratch!

    Parameters
    ----------
    prefix : str, optional
        path prefix to temporary folder, by default None

    Yields
    ------
    temp_dir : str
        name of the temporary directory
    """
    mycwd = os.getcwd()
    try:
        temp_dir = tempfile.mkdtemp(prefix=prefix)
        os.chdir(temp_dir)
        yield temp_dir
    finally:
        os.chdir(mycwd)
        shutil.rmtree(temp_dir)
        


import json
import re
from ase.db import connect
import numpy as np
import pandas as pd
from tqdm import tqdm

from architector import io_ptable
# import pathlib -> Useful for listing jsons.

def serialize_json_dict(indict):
    """serialize_json_dict ase json serialization routine

    Parameters
    ----------
    indict : dict
        json dictionary

    Returns
    -------
    ss : str
        json serialized version of the 
    """
    ss = '{'
    count = 1
    ids = []
    for key,val in tqdm(indict.items(),total=len(indict)):
        if count % 500 == 0:
            print(count)
        tlines = '"{}"'.format(count) + ': '+ '{\n'
        for skey,sval in val.items():
            if isinstance(sval,str):
                sval = '"{}"'.format(sval)
            tlines += '"{}"'.format(skey)+': '+str(sval)+',\n'
        tlines = tlines.strip().strip(',') + '},\n'
        tlines = re.sub("'",'"',tlines)
        tlines = re.sub('True','true',tlines)
        tlines = re.sub('False','false',tlines)
        if 'nan' in tlines:
            print('nan error')
        else:
            ids.append(count)
            count += 1
            ss += tlines
    ss += '"ids": '+str(ids)+',\n'
    ss += '"nextid": '+str(count)+'}'
    return ss


def merge_JsonFiles(filenamelist,outfname='compiled.json'):
    """jsons = [str(x) for x in p.glob('architector*json')]
    compiled_json_name = 'compiled.json'
    merge_JsonFiles(jsons,compiled_json_name)"""
    result = dict()
    ids = []
    count = 1
    for i,f1 in tqdm(enumerate(filenamelist),total=len(filenamelist)):
        newdata = dict()
        try:
            with open(f1, 'r') as infile:
                data = json.load(infile)
            for key,val in data.items():
                if (key != 'ids') and (key != 'nextid'):
                    newdata[str(count)] = val
                    ids.append(count)
                    count +=1
            result.update(newdata)
        except Exception as e:
            print('Filename: {} Failed'.format(f1))

    with open(outfname, 'w') as output_file:
        ss = serialize_json_dict(result)
        output_file.write(ss)


def convert_arrays_to_npz(dbname, prefix, mindist_cutoff=0.5,
    return_symbols=False, max_force=300, return_df=False):
    """load arrays load ase database into hippynn database arrays
    example : convert_arrays_to_npz('compiled.json','xtbdataset')

    Parameters
    ----------
    filename : str
        filename or path of database to convert
    prefix : str, optional
        prefix for output numpy arrays, by default None
    mindist_cutoff : float, optional
        minimum distance cutoff, default 0.5 Angstroms
    return_symbols : bool, optional
        return the symbols of all structures. Default False
    max_force : float, optional
        cutoff to remove max forces. Default 300 ev/A
    return_df : bool, optional,
        return dataframe with all atoms information, Default False
    """
    try:
        db = connect(dbname)
    except Exception as e:
        print(e)
    record_list = []
    symbols = []
    max_n_atom = 0
    total = None
    with open(dbname, 'r') as f:
        for line in f:
            pass
        last_line = line
    print(last_line)
    total = int(last_line.split()[1].split('}')[0])
    any_pbc = False
    for row in tqdm(db.select(),total=total):
        is_pbc = False
        if np.any(row.pbc):
            is_pbc = True
            any_pbc = True
        try:
            syms,counts=np.unique(row.symbols,return_counts=True)
            result_dict = {
                'atoms':row.numbers,
                'xyz': row.positions,
                'cell': row.cell,
                'is_pbc':is_pbc,
                'force': row.forces,
                'energy': row.energy,
                'atomization_energy':row.energy -  np.sum([io_ptable.xtb_single_atom_ref_es[sym]*counts[i] for i,sym in enumerate(syms)]),
                'uid': row.unique_id,
                'relaxed': row.relaxed,
                'geo_step': row.geo_step
            }
            distmat = row.toatoms().get_all_distances() + np.eye(len(row.numbers))* mindist_cutoff*2
            maxforce = np.max(np.linalg.norm(row.forces,axis=1))
            if (distmat.min() > mindist_cutoff) and (maxforce < max_force): # Hard distance cutoff, forces cutoff
                record_list.append(result_dict)
                if row.natoms > max_n_atom:
                    max_n_atom = row.natoms
                if return_symbols:
                    symbols += row.symbols
        except:
            pass
    n_record = len(record_list)
    # Sort the list base on number of atoms
    record_list.sort(key=lambda rec: len(rec['atoms']))
    # Save the record uid from the ase db object names.
    xyz_array = np.zeros([n_record, max_n_atom, 3])
    force_array = np.zeros([n_record, max_n_atom, 3])
    atom_z_array = np.zeros([n_record, max_n_atom])
    if any_pbc:
        cell_array = np.zeros([n_record, 3, 3])
        pbc = True
    else:
        cell_array = np.zeros([n_record, 3, 3])
        pbc = False
    energy_array = np.array([record['energy'] for record in record_list])
    atomization_array = np.array([record['atomization_energy'] for record in record_list])
    for i, record in enumerate(record_list):
        natom = len(record['atoms'])
        xyz_array[i, :natom, :] = record['xyz']
        force_array[i, :natom, :] = record['force']
        atom_z_array[i,:natom] = record['atoms']
        if pbc:
            cell_array[i, :, :] = record['cell']
    np.save("data-" + prefix + 'atomization_energy.npy', atomization_array) 
    np.save("data-" + prefix + 'energy.npy', energy_array)        
    np.save("data-" + prefix + 'R.npy', xyz_array)
    np.save("data-" + prefix + 'force.npy', force_array)
    np.save("data-" + prefix + 'Z.npy', atom_z_array.astype('int'))
    if return_symbols:
        return symbols
    if return_df:
        return pd.DataFrame(record_list)


"""
Py3Dmol install: (works in both Python2/3 conda environments)
conda install -c conda-forge py3dmol 
Some Documentation: https://pypi.org/project/py3Dmol/
3DMol.js backend: http://3dmol.csb.pitt.edu/index.html

Visualization routine for handling jupyter notebook use of architector.

Developed by Michael Taylor
"""
import math as m
import numpy as np
import ase

import py3Dmol

import architector
from architector.io_molecule import convert_io_molecule
import architector.io_ptable as io_ptable

def type_convert(structures):
    """Handle multiple types of structures passed. List of xyz, mol2 files,
    or list of xyz, mol2strings.

    Parameters
    ----------
    structures : list
        Structures you want visualized: can either be a list or individual:
        mol2 strings, mol2 files, xyz strings, xyz files, or mol3D objects
    """
    outlist = []
    if isinstance(structures,str):
        structures = [convert_io_molecule(structures)]
    elif isinstance(structures,(ase.atoms.Atoms,architector.io_molecule.Molecule)):
        structures = [convert_io_molecule(structures)]
    elif isinstance(structures,dict):
        try:
            structures = [val['mol2string'] for key,val in structures.items()]
        except:
            raise ValueError('Not recognized type for this dictionary to visualize!')
    else: # Convert other array-like arguments to a list.
        structures = list(structures)
    if isinstance(structures,list):
        for i,x in enumerate(structures):
            try: 
                mol = convert_io_molecule(x)
                outlist.append(mol)
            except:
                raise ValueError('Not Recognized Structure Type for index: ' +str(i))
    return outlist
                
            
def view_structures(structures,w=200,h=200,columns=4,representation='stick',labelsize=12,
                 labels=False, labelinds=False, vector=None):
    """
    py3Dmol view atoms object(s)
    xyz_names = xyz files that will be rendered in a tiled format in jupyter (list,str)
    w = width of frame (or subframes) in pixels (int)
    h = height of frame (or subframes) in pixels (int)
    cols = number of columns in subframe (int)
    representation = how the molecule will be viewed (str)
    labelsize = size of the data label (in Points) (int)
    labels = turn labels on/off (bool)
    labelinds = whether to add the indices as text objects to the visualized molecule.
    vector = {'start': {'x':-10.0, 'y':0.0, 'z':0.0}, 'end': {'x':-10.0, 'y':0.0, 'z':10.0},
              'radius':2,'color':'red'}
    """
    mols = type_convert(structures)
    if len(mols) == 1:
        view_ats = py3Dmol.view(width=w,height=h)
        mol = mols[0]
        if isinstance(labels,str):
            label = labels
        elif isinstance(labels,list):
            label = labels[0]
        elif isinstance(labels,bool):
            if labels:
                label = mol.ase_atoms.get_chemical_formula()
            else:
                label = False
        metal_ind = [i for i,x in enumerate(mol.ase_atoms) if (x.symbol in io_ptable.all_metals)]
        if len(metal_ind) > 0 : # Take advantage of empty list
            label_posits = mol.ase_atoms.positions[metal_ind].flatten()
        else:
            label_posits = mol.ase_atoms.get_center_of_mass().flatten()  # Put it at the geometric center of the molecule.
        coords = mol.write_mol2('tmp.mol2', writestring=True)
        view_ats.addModel(coords.replace('un','1'),'mol2') # Add the molecule
        view_ats.setStyle({representation:{'colorscheme':'Jmol'}}) 
        if label:
            view_ats.addLabel("{}".format(label), {'position':{'x':'{}'.format(label_posits[0]),
                  'y':'{}'.format(label_posits[1]),'z':'{}'.format(label_posits[2])},
                  'backgroundColor':"'black'",'backgroundOpacity':'0.3',
                  'fontOpacity':'1', 'fontSize':'{}'.format(labelsize),
                  'fontColor':"white",'inFront':'true'})
        if labelinds:
            inds = [x for x in range(len(mol.ase_atoms))]
            for i in inds:
                atom_posit = mol.ase_atoms.positions[i]
                view_ats.addLabel("{}".format(i), {'position':{'x':'{}'.format(atom_posit[0]),
                  'y':'{}'.format(atom_posit[1]),'z':'{}'.format(atom_posit[2])},
                  'backgroundColor':"'black'",'backgroundOpacity':'0.4',
                  'fontOpacity':'1', 'fontSize':'{}'.format(labelsize),
                  'fontColor':"white",'inFront':'true'})
        if vector:
            view_ats.addArrow(vector)
        view_ats.zoomTo()
        view_ats.show()
    elif len(mols) < 50:
        rows = int(m.ceil(float(len(mols))/columns))
        w = w*columns
        h = h*rows 
        # Initialize Layout
        view_ats = py3Dmol.view(width=w,height=h,linked=False,viewergrid=(rows,columns))
        # Check for labels and populate
        if isinstance(labels,bool):
            if labels:
                label = [x.ase_atoms.get_chemical_formula() for x in mols]
            else:
                label = []
        elif isinstance(labels,list) or isinstance(labels,np.ndarray):
            if (len(labels) != len(mols)):
                print('Wrong amount of labels passed, defaulting to chemical formulas.')
                label = [x.ase_atoms.get_chemical_formula() for x in mols]
            else: # Force them all to be strings. 
                label = [str(x) for x in labels]
        else:
            raise ValueError('What sort of labels are wanting? Not recognized.')
        x,y = 0,0 # Subframe position
        for i,mol in enumerate(mols):
            metal_inds = [i for i,x in enumerate(mol.ase_atoms) if (x.symbol in io_ptable.all_metals)]
            if len(metal_inds) > 0 : # Take advantage of empty list
                label_posits = mol.ase_atoms.positions[metal_inds[0]].flatten()
            else:
                label_posits = mol.ase_atoms.get_center_of_mass().flatten()  # Put it at the geometric center of the molecule.
            coords = mol.write_mol2('tmp.mol2', writestring=True)
            view_ats.addModel(coords.replace('un','1'),'mol2',viewer=(x,y))
            view_ats.setStyle({representation:{'colorscheme':'Jmol'}},viewer=(x,y))
            if len(label) > 0:
                view_ats.addLabel("{}".format(label[i]), {'position':{'x':'{}'.format(label_posits[0]),
                    'y':'{}'.format(label_posits[1]),'z':'{}'.format(label_posits[2])},
                    'backgroundColor':"'black'",'backgroundOpacity':'0.5',
                    'fontOpacity':'1','fontSize':'{}'.format(labelsize),
                    'fontColor':"white",'inFront':'true',}, viewer=(x,y))
            if labelinds:
                inds = [x for x in range(len(mol.ase_atoms))]
                for j in inds:
                    atom_posit = mol.ase_atoms.positions[j]
                    view_ats.addLabel("{}".format(j), {'position':{'x':'{}'.format(atom_posit[0]),
                    'y':'{}'.format(atom_posit[1]),'z':'{}'.format(atom_posit[2])},
                    'backgroundColor':"'black'",'backgroundOpacity':'0.4',
                    'fontOpacity':'1', 'fontSize':'{}'.format(int(labelsize)),
                    'fontColor':"white", 'inFront':'true'}, viewer=(x,y))
            if vector:
                view_ats.addArrow(vector)
            view_ats.zoomTo(viewer=(x,y))
            if y+1 < columns: # Fill in columns
                y+=1
            else:
                x+=1
                y=0
        view_ats.show()
    else: 
        raise ValueError('Warning. Passing this many structures WILL cause your kernel to crash.')
